# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Resources module."""
from invenio_vocabularies.resources.schema import L10NString, VocabularyL10Schema

__all__ = (
    "VocabularyL10Schema",
    "L10NString",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Test funder schema."""

import pytest
from marshmallow import ValidationError

from invenio_vocabularies.contrib.funders.schema import (
    FunderRelationSchema,
    FunderSchema,
)


def test_valid_full(app, funder_full_data):
    loaded = FunderSchema().load(funder_full_data)
    funder_full_data["pid"] = funder_full_data.pop("id")
    assert funder_full_data == loaded


def test_valid_minimal(app):
    data = {
        "id": "01ggx4157",
        "name": "Test funder",
    }
    assert {
        "pid": "01ggx4157",
        "name": "Test funder",
    } == FunderSchema().load(data)


def test_invalid_no_name():
    invalid_no_name = {
        "id": "01ggx4157",
        "identifiers": [
            {
                "identifier": "000000012156142X",
                "scheme": "isni",
            },
            {
                "identifier": "grid.9132.9",
                "scheme": "grid",
            },
        ],
        "title": {
            "en": "European Organization for Nuclear Research",
            "fr": "Organisation européenne pour la recherche nucléaire",
        },
        "country": "CH",
    }
    with pytest.raises(ValidationError):
        data = FunderSchema().load(invalid_no_name)


def test_invalid_empty(app):
    data = {
        "id": "01ggx4157",
        "name": "",
    }
    with pytest.raises(ValidationError):
        data = FunderSchema().load(data)

    data = {
        "id": "",
        "name": "Test funder",
    }
    with pytest.raises(ValidationError):
        data = FunderSchema().load(data)


def test_invalid_empty_funder():
    invalid_empty = {}
    with pytest.raises(ValidationError):
        data = FunderSchema().load(invalid_empty)


def test_invalid_country():
    invalid_country = {"id": "01ggx4157", "name": "Test funder", "country": 1}
    with pytest.raises(ValidationError):
        data = FunderSchema().load(invalid_country)


#
# FunderRelationSchema
#


def test_valid_id():
    valid_id = {
        "id": "test",
    }
    assert valid_id == FunderRelationSchema().load(valid_id)


def test_valid_name():
    valid_name = {"name": "Funder One"}
    assert valid_name == FunderRelationSchema().load(valid_name)


def test_valid_both_id_name():
    valid_id_name = {"id": "test", "name": "Funder One"}
    assert valid_id_name == FunderRelationSchema().load(valid_id_name)


def test_invalid_empty_relation():
    invalid_empty = {}
    with pytest.raises(ValidationError):
        data = FunderRelationSchema().load(invalid_empty)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""DataStreams tests."""

import json
import zipfile
from pathlib import Path

import pytest

from invenio_vocabularies.datastreams.factories import DataStreamFactory


@pytest.fixture(scope="module")
def vocabulary_config():
    """Parsed vocabulary configuration."""
    return {
        "transformers": [{"type": "test"}],
        "readers": [
            {
                "type": "test",
                "args": {
                    "origin": [1, -1],
                },
            }
        ],
        "writers": [{"type": "test"}],
    }


def test_base_datastream(app, vocabulary_config):
    datastream = DataStreamFactory.create(
        readers_config=vocabulary_config["readers"],
        transformers_config=vocabulary_config.get("transformers"),
        writers_config=vocabulary_config["writers"],
    )

    stream_iter = datastream.process()
    valid = next(stream_iter)
    assert valid.entry == 2
    assert not valid.errors

    invalid = next(stream_iter)
    assert invalid.entry == -1
    assert "TestTransformer: Value cannot be negative" in invalid.errors


def test_base_datastream_fail_on_write(app, vocabulary_config):
    custom_config = dict(vocabulary_config)
    custom_config["writers"].append(
        {
            "type": "fail",
            "args": {"fail_on": 2},  # 2 means 1 as entry cuz transformers sums 1
        }
    )

    datastream = DataStreamFactory.create(
        readers_config=vocabulary_config["readers"],
        transformers_config=vocabulary_config.get("transformers"),
        writers_config=vocabulary_config["writers"],
    )

    stream_iter = datastream.process()
    invalid_wr = next(stream_iter)
    assert invalid_wr.entry == 2  # entry got transformed
    assert "FailingTestWriter: 2 value found." in invalid_wr.errors

    # failed on the previous but can process the next
    invalid_tr = next(stream_iter)
    assert invalid_tr.entry == -1
    assert "TestTransformer: Value cannot be negative" in invalid_tr.errors


@pytest.fixture(scope="function")
def zip_file(json_list):
    """Creates a Zip file with three files inside.

    The first file should return two json elements.
    The second file should fail.
    The third file should return two json elements.
    """

    def _correct_file(archive, idx):
        correct_file = Path(f"correct_{idx}.json")
        with open(correct_file, "w") as file:
            json.dump(json_list, file)
        archive.write(correct_file)
        correct_file.unlink()

    filename = Path("reader_test.zip")
    with zipfile.ZipFile(file=filename, mode="w") as archive:
        _correct_file(archive, 1)
        errored_file = Path("errored.json")
        with open(errored_file, "w") as file:
            file.write(  # to dump incorrect json format
                # missing comma and closing bracket
                '[{"test": {"inner": "value"}{"test": {"inner": "value"}}]'
            )
        archive.write(errored_file)
        _correct_file(archive, 2)
        errored_file.unlink()

    yield filename

    filename.unlink()  # delete created file


def test_piping_readers(app, zip_file, json_element):
    ds_config = {
        "readers": [
            {
                "type": "zip",
                "args": {
                    "origin": "reader_test.zip",
                    "regex": ".json$",
                },
            },
            {"type": "json"},
        ],
        "writers": [{"type": "test"}],
    }

    datastream = DataStreamFactory.create(
        readers_config=ds_config["readers"],
        writers_config=ds_config["writers"],
    )
    expected_errors = [
        "ZipReader.read: Cannot decode JSON file errored.json: Expecting ',' delimiter: line 1 column 29 (char 28)"  # noqa
    ]

    iter = datastream.process()
    for count, entry in enumerate(iter, start=1):
        if count != 3:
            assert entry.entry == json_element
        else:
            # assert the second file fails
            assert entry.entry.name == "errored.json"
            assert entry.errors == expected_errors

    assert count == 5  # 2 good + 1 bad + 2 good



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Affiliations API tests."""

from copy import deepcopy
from functools import partial

import pytest
from invenio_indexer.api import RecordIndexer
from invenio_search import current_search_client
from jsonschema import ValidationError as SchemaValidationError

from invenio_vocabularies.contrib.affiliations.api import Affiliation


@pytest.fixture()
def search_get():
    """Get a document from an index."""
    return partial(current_search_client.get, Affiliation.index._name)


@pytest.fixture()
def indexer():
    """Indexer instance with correct Record class."""
    return RecordIndexer(
        record_cls=Affiliation,
        record_to_index=lambda r: (r.__class__.index._name, "_doc"),
    )


@pytest.fixture()
def example_affiliation(db, affiliation_full_data):
    """Example affiliation."""
    api_affiliation = deepcopy(affiliation_full_data)
    pid = api_affiliation.pop("id")
    aff = Affiliation.create(api_affiliation, pid=pid)
    aff.commit()
    db.session.commit()
    return aff


def test_affiliation_schema_validation(app, db, example_affiliation):
    """Affiliation schema validation."""
    # valid data
    aff = example_affiliation

    assert aff.schema == "local://affiliations/affiliation-v1.0.0.json"
    assert aff.pid
    assert aff.id

    # invalid data
    examples = [
        # title are objects of key/string.
        {"id": "cern", "name": "cern", "title": "not a dict"},
        {"id": "cern", "name": "cern", "title": {"en": 123}},
        # identifiers are objects of key/string.
        {"id": "cern", "name": "cern", "identifiers": "03yrm5c26"},
        {"id": "cern", "name": "cern", "identifiers": ["03yrm5c26"]},
        {"id": "cern", "name": "cern", "identifiers": {"03yrm5c26"}},
        # name must be a string
        {"id": "cern", "name": 123},
        # acronym must be a string
        {"id": "cern", "name": "cern", "acronym": 123},
    ]

    for ex in examples:
        pytest.raises(SchemaValidationError, Affiliation.create, ex)


def test_affiliation_indexing(
    app, db, search, example_affiliation, indexer, search_get
):
    """Test indexing of an affiliation."""
    # Index document in ES
    assert indexer.index(example_affiliation)["result"] == "created"

    # Retrieve document from ES
    data = search_get(id=example_affiliation.id)

    # Loads the ES data and compare
    aff = Affiliation.loads(data["_source"])
    assert aff == example_affiliation
    assert aff.id == example_affiliation.id
    assert aff.revision_id == example_affiliation.revision_id
    assert aff.created == example_affiliation.created
    assert aff.updated == example_affiliation.updated


def test_affiliation_pid(app, db, example_affiliation):
    """Test affiliation pid creation."""
    aff = example_affiliation

    assert aff.pid.pid_value == "cern"
    assert Affiliation.pid.resolve("cern")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Transformers module."""

from abc import ABC, abstractmethod
from collections import defaultdict

from lxml import etree

from .errors import TransformerError


class BaseTransformer(ABC):
    """Base transformer."""

    @abstractmethod
    def apply(self, stream_entry, *args, **kwargs):
        """Applies the transformation to the entry.

        :returns: A StreamEntry. The transformed entry.
                  Raises TransformerError in case of errors.
        """
        pass


class XMLTransformer(BaseTransformer):
    """XML transformer."""

    @classmethod
    def _xml_to_etree(cls, xml):
        """Converts XML to a lxml etree."""
        return etree.HTML(xml)

    @classmethod
    def _etree_to_dict(cls, tree):
        d = {tree.tag: {} if tree.attrib else None}
        children = list(tree)
        if children:
            dd = defaultdict(list)
            for dc in map(cls._etree_to_dict, children):
                for k, v in dc.items():
                    dd[k].append(v)
            d = {tree.tag: {k: v[0] if len(v) == 1 else v for k, v in dd.items()}}
        if tree.attrib:
            d[tree.tag].update(("@" + k, v) for k, v in tree.attrib.items())
        if tree.text:
            text = tree.text.strip()
            if children or tree.attrib:
                if text:
                    d[tree.tag]["#text"] = text
            else:
                d[tree.tag] = text
        return d

    def apply(self, stream_entry, **kwargs):
        """Applies the transformation to the stream entry.

        Requires the root element to be named "record".
        """
        xml_tree = self._xml_to_etree(stream_entry.entry)
        record = self._etree_to_dict(xml_tree)["html"]["body"].get("record")

        if not record:
            raise TransformerError(f"Record not found in XML entry.")

        stream_entry.entry = record
        return stream_entry



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Names module."""

from .resources import NamesResource, NamesResourceConfig
from .services import NamesService, NamesServiceConfig

__all__ = (
    "NamesResource",
    "NamesResourceConfig" "NamesService",
    "NamesServiceConfig",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Awards datastreams tests."""

from copy import deepcopy

import pytest
from invenio_access.permissions import system_identity

from invenio_vocabularies.contrib.awards.api import Award
from invenio_vocabularies.contrib.awards.datastreams import (
    AwardsServiceWriter,
    OpenAIREProjectTransformer,
)
from invenio_vocabularies.datastreams import StreamEntry
from invenio_vocabularies.datastreams.errors import WriterError


@pytest.fixture(scope="function")
def dict_award_entry():
    return StreamEntry(
        {
            "acronym": "TA",
            "code": "0751743",
            "enddate": "2010-09-30",
            "funding": [
                {
                    "funding_stream": {
                        "description": "Directorate for Geosciences - Division of "
                        "Ocean Sciences",
                        "id": "NSF::GEO/OAD::GEO/OCE",
                    },
                    "jurisdiction": "US",
                    "name": "National Science Foundation",
                    "shortName": "NSF",
                }
            ],
            "h2020programme": [],
            "id": "40|nsf_________::3eb1b4f6d6e251a19f9fdeed2aab88d8",
            "openaccessmandatefordataset": False,
            "openaccessmandateforpublications": False,
            "startdate": "2008-04-01",
            "subject": ["Oceanography"],
            "title": "Test title",
            "websiteurl": "https://test.com",
        }
    )


@pytest.fixture(scope="function")
def dict_award_entry_ec():
    """Full award data."""
    return StreamEntry(
        {
            "acronym": "TS",
            "code": "129123",
            "enddate": "2025-12-31",
            "funding": [
                {
                    "funding_stream": {
                        "description": "Test stream",
                        "id": "TST::test::test",
                    },
                    "jurisdiction": "GR",
                    "name": "Test Name",
                    "shortName": "TST",
                }
            ],
            "h2020programme": [],
            "id": "40|corda__h2020::000000000000000000",
            "openaccessmandatefordataset": False,
            "openaccessmandateforpublications": False,
            "startdate": "2008-04-01",
            "subject": ["Oceanography"],
            "title": "Test title",
            "websiteurl": "https://test.com",
        }
    )


@pytest.fixture(scope="function")
def expected_from_award_json():
    return {
        "id": "021nxhr62::0751743",
        "identifiers": [{"identifier": "https://test.com", "scheme": "url"}],
        "number": "0751743",
        "title": {"en": "Test title"},
        "funder": {"id": "021nxhr62"},
        "acronym": "TA",
        "program": "GEO/OAD",
    }


@pytest.fixture(scope="function")
def expected_from_award_json_ec():
    return {
        "id": "00k4n6c32::129123",
        "identifiers": [
            {"identifier": "https://cordis.europa.eu/projects/129123", "scheme": "url"}
        ],
        "number": "129123",
        "title": {"en": "Test title"},
        "funder": {"id": "00k4n6c32"},
        "acronym": "TS",
        "program": "test",
    }


def test_awards_transformer(app, dict_award_entry, expected_from_award_json):
    transformer = OpenAIREProjectTransformer()
    assert expected_from_award_json == transformer.apply(dict_award_entry).entry


def test_awards_service_writer_create(
    app, search_clear, example_funder_ec, award_full_data
):
    awards_writer = AwardsServiceWriter()
    award_rec = awards_writer.write(StreamEntry(award_full_data))
    award_dict = award_rec.entry.to_dict()

    award_full_data["funder"]["name"] = example_funder_ec["name"]
    assert dict(award_dict, **award_full_data) == award_dict

    # not-ideal cleanup
    award_rec.entry._record.delete(force=True)


def test_awards_funder_id_not_exist(
    app,
    search_clear,
    example_funder,
    example_funder_ec,
    award_full_data_invalid_id,
):
    awards_writer = AwardsServiceWriter()
    with pytest.raises(WriterError) as err:
        awards_writer.write(StreamEntry(award_full_data_invalid_id))
    expected_error = [
        {
            "InvalidRelationValue": "Invalid value {funder_id}.".format(
                funder_id=award_full_data_invalid_id.get("funder").get("id")
            )
        }
    ]

    assert expected_error in err.value.args


def test_awards_funder_id_not_exist_no_funders(
    app, search_clear, award_full_data_invalid_id
):
    awards_writer = AwardsServiceWriter()
    with pytest.raises(WriterError) as err:
        awards_writer.write(StreamEntry(award_full_data_invalid_id))
    expected_error = [
        {
            "InvalidRelationValue": "Invalid value {funder_id}.".format(
                funder_id=award_full_data_invalid_id.get("funder").get("id")
            )
        }
    ]

    assert expected_error in err.value.args


def test_awards_transformer_ec_functionality(
    app,
    search_clear,
    dict_award_entry,
    dict_award_entry_ec,
    expected_from_award_json,
    expected_from_award_json_ec,
):
    transformer = OpenAIREProjectTransformer()
    assert expected_from_award_json == transformer.apply(dict_award_entry).entry
    assert expected_from_award_json_ec == transformer.apply(dict_award_entry_ec).entry


def test_awards_service_writer_duplicate(
    app, search_clear, example_funder_ec, award_full_data
):
    writer = AwardsServiceWriter()
    award_rec = writer.write(stream_entry=StreamEntry(award_full_data))
    Award.index.refresh()  # refresh index to make changes live
    with pytest.raises(WriterError) as err:
        writer.write(stream_entry=StreamEntry(award_full_data))

    expected_error = [f"Vocabulary entry already exists: {award_full_data}"]
    assert expected_error in err.value.args

    # not-ideal cleanup
    award_rec.entry._record.delete(force=True)


def test_awards_service_writer_update_existing(
    app, search_clear, example_funder_ec, award_full_data, service
):
    # create vocabulary
    writer = AwardsServiceWriter(update=True)
    orig_award_rec = writer.write(stream_entry=StreamEntry(award_full_data))
    Award.index.refresh()  # refresh index to make changes live
    # update vocabulary
    updated_award = deepcopy(award_full_data)
    updated_award["title"] = {"en": "New Test title"}
    # check changes vocabulary
    _ = writer.write(stream_entry=StreamEntry(updated_award))
    award_rec = service.read(system_identity, orig_award_rec.entry.id)
    award_dict = award_rec.to_dict()

    updated_award["funder"]["name"] = example_funder_ec["name"]
    # needed while the writer resolves from ES
    assert _.entry.id == orig_award_rec.entry.id
    assert dict(award_dict, **updated_award) == award_dict

    # not-ideal cleanup
    award_rec._record.delete(force=True)


def test_awards_service_writer_update_non_existing(
    app, search_clear, example_funder_ec, award_full_data, service
):
    # vocabulary item not created, call update directly
    updated_award = deepcopy(award_full_data)
    updated_award["title"] = {"en": "New Test title"}
    # check changes vocabulary
    writer = AwardsServiceWriter(update=True)
    award_rec = writer.write(stream_entry=StreamEntry(updated_award))
    award_rec = service.read(system_identity, award_rec.entry.id)
    award_dict = award_rec.to_dict()

    updated_award["funder"]["name"] = example_funder_ec["name"]
    assert dict(award_dict, **updated_award) == award_dict

    # not-ideal cleanup
    award_rec._record.delete(force=True)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Resources layer tests."""

import pytest

from invenio_vocabularies.records.api import Vocabulary


#
# Fixtures
#
@pytest.fixture()
def example_data():
    """Example data for records."""
    return [
        {
            "id": "cc-by",
            "title": {"en": "Creative Commons Attribution"},
            "type": "licenses",
        },
        {"id": "cc0", "title": {"en": "Creative Commons Zero"}, "type": "licenses"},
    ]


@pytest.fixture()
def example_records(db, identity, service, example_data):
    """Create some example records."""
    service.create_type(identity, "licenses", "lic")
    records = []
    for data in example_data:
        records.append(service.create(identity, data))
    Vocabulary.index.refresh()
    return records


@pytest.fixture()
def prefix():
    """API prefix."""
    return "/vocabularies/licenses/"


#
# Tests
#
def test_get(client, example_records, h, prefix):
    """Test the endpoint to retrieve a single item."""
    id_ = example_records[0].id

    res = client.get(f"{prefix}{id_}", headers=h)
    assert res.status_code == 200
    assert res.json["id"] == id_
    # Test links
    assert res.json["links"] == {
        "self": "https://127.0.0.1:5000/api/vocabularies/licenses/cc-by"
    }


def test_not_found(client, example_records, h, prefix):
    """Test not found."""
    id_ = example_records[0].id
    # invalid id
    res = client.get(
        f"{prefix}invalid-id",
        headers=h,
    )
    assert res.status_code == 404
    # invalid type (wrong type for pid)
    res = client.get(
        f"/vocabularies/invalid/{id_}",
        headers=h,
    )
    assert res.status_code == 404
    # trailing slash on search
    res = client.get(
        f"{prefix}/",
        headers=h,
    )
    assert res.status_code == 404
    # invalid type (not existing)
    res = client.get(
        f"/vocabularies/invalid",
        headers=h,
    )
    assert res.status_code == 404



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2021 Northwestern University.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary subjects."""

from invenio_records.dumpers import SearchDumper
from invenio_records.dumpers.indexedat import IndexedAtDumperExt
from invenio_records_resources.factories.factory import RecordTypeFactory

from ...records.pidprovider import PIDProviderFactory
from ...records.systemfields import BaseVocabularyPIDFieldContext
from ...services.permissions import PermissionPolicy
from .config import SubjectsSearchOptions, service_components
from .schema import SubjectSchema

record_type = RecordTypeFactory(
    "Subject",
    # Data layer
    pid_field_kwargs={
        "create": False,
        "provider": PIDProviderFactory.create(pid_type="sub"),
        "context_cls": BaseVocabularyPIDFieldContext,
    },
    schema_version="1.0.0",
    schema_path="local://subjects/subject-v1.0.0.json",
    record_dumper=SearchDumper(
        extensions=[
            IndexedAtDumperExt(),
        ]
    ),
    # Service layer
    service_id="subjects",
    service_schema=SubjectSchema,
    search_options=SubjectsSearchOptions,
    service_components=service_components,
    permission_policy_cls=PermissionPolicy,
    # Resource layer
    endpoint_route="/subjects",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 Northwestern University.
# Copyright (C) 2021-2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Subjects schema."""

from invenio_i18n import lazy_gettext as _
from marshmallow_utils.fields import SanitizedUnicode

from ...services.schema import BaseVocabularySchema, ContribVocabularyRelationSchema


class SubjectSchema(BaseVocabularySchema):
    """Service schema for subjects."""

    # id in BaseRecordSchema is not required, but I don't see why it shouldn't
    # be, while scheme and subject are required. So I am making it required
    # here.
    id = SanitizedUnicode(required=True)
    scheme = SanitizedUnicode(required=True)
    subject = SanitizedUnicode(required=True)


class SubjectRelationSchema(ContribVocabularyRelationSchema):
    """Schema to define an optional subject relation in another schema."""

    ftf_name = "subject"
    parent_field_name = "subjects"
    subject = SanitizedUnicode()
    scheme = SanitizedUnicode()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 Northwestern University.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Subject API tests."""

from functools import partial

import pytest
from invenio_indexer.api import RecordIndexer
from invenio_search import current_search_client

from invenio_vocabularies.contrib.subjects.api import Subject


@pytest.fixture()
def search_get():
    """Get a document from an index."""
    return partial(current_search_client.get, Subject.index._name)


@pytest.fixture()
def indexer():
    """Indexer instance with correct Record class."""
    return RecordIndexer(
        record_cls=Subject,
        record_to_index=lambda r: (r.__class__.index._name, "_doc"),
    )


@pytest.fixture()
def example_subject(db, subject_full_data):
    """Example subject."""
    subj = Subject.create(subject_full_data)
    Subject.pid.create(subj)
    subj.commit()
    db.session.commit()
    return subj


def test_subject_indexing(app, db, search, example_subject, indexer, search_get):
    """Test indexing of a subject."""
    # Index document in ES
    assert indexer.index(example_subject)["result"] == "created"

    # Retrieve document from ES
    data = search_get(id=example_subject.id)

    # Loads the ES data and compare
    subj = Subject.loads(data["_source"])
    assert subj == example_subject
    assert subj.id == example_subject.id
    assert subj.revision_id == example_subject.revision_id
    assert subj.created == example_subject.created
    assert subj.updated == example_subject.updated


def test_subject_pid(app, db, example_subject):
    """Test subject pid creation."""
    subj = example_subject

    assert subj.pid.pid_value == "https://id.nlm.nih.gov/mesh/D000001"
    assert subj.pid.pid_type == "sub"
    assert Subject.pid.resolve("https://id.nlm.nih.gov/mesh/D000001")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary affiliations configuration."""

from flask import current_app
from invenio_i18n import lazy_gettext as _
from invenio_records_resources.services import SearchOptions
from invenio_records_resources.services.records.components import DataComponent
from invenio_records_resources.services.records.params import SuggestQueryParser
from werkzeug.local import LocalProxy

from ...services.components import PIDComponent

affiliation_schemes = LocalProxy(
    lambda: current_app.config["VOCABULARIES_AFFILIATION_SCHEMES"]
)


class AffiliationsSearchOptions(SearchOptions):
    """Search options."""

    suggest_parser_cls = SuggestQueryParser.factory(
        fields=[
            "name^100",
            "acronym^20",
            "title.*^5",
            "title.*._2gram",
            "title.*._3gram",
        ],
    )

    sort_default = "bestmatch"

    sort_default_no_query = "name"

    sort_options = {
        "bestmatch": dict(
            title=_("Best match"),
            fields=["_score"],  # ES defaults to desc on `_score` field
        ),
        "name": dict(
            title=_("Name"),
            fields=["name_sort"],
        ),
        "newest": dict(
            title=_("Newest"),
            fields=["-created"],
        ),
        "oldest": dict(
            title=_("Oldest"),
            fields=["created"],
        ),
    }


service_components = [
    # Order of components are important!
    DataComponent,
    PIDComponent,
]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Tasks REST API tests."""

import json
from unittest.mock import patch

import pytest

from invenio_vocabularies.datastreams.datastreams import StreamEntry
from invenio_vocabularies.datastreams.readers import BaseReader
from invenio_vocabularies.datastreams.transformers import BaseTransformer
from invenio_vocabularies.datastreams.writers import BaseWriter


class TestReader(BaseReader):
    """Test reader."""

    # needs implementation due to abstract decorator in the base class
    def _iter(self, fp, *args, **kwargs):
        pass

    def read(self, item=None, *args, **kwargs):
        """Reads from item or opens the file descriptor from origin."""
        yield 1


class TestTransformer(BaseTransformer):
    """Test transformer."""

    # no need for self since it is patched
    def apply(stream_entry, *args, **kwargs):
        """Forwards entry."""
        return stream_entry


class TestWriter(BaseWriter):
    """Test writer."""

    # no need for self since it is patched
    def write(stream_entry, *args, **kwargs):
        """NOP write."""
        pass


@pytest.fixture(scope="module")
def app_config(app_config):
    """Mimic an instance's configuration."""
    app_config["VOCABULARIES_DATASTREAM_READERS"] = {"test": TestReader}
    app_config["VOCABULARIES_DATASTREAM_TRANSFORMERS"] = {"test": TestTransformer}
    app_config["VOCABULARIES_DATASTREAM_WRITERS"] = {"test": TestWriter}

    return app_config


def test_task_creation(app, client_with_credentials, h):
    client = client_with_credentials
    task_config = {
        "readers": [{"type": "test", "args": {"origin": "somewhere"}}],
        "transformers": [{"type": "test"}],
        "writers": [{"type": "test"}],
    }

    with patch.object(
        TestReader, "read", side_effect=TestReader.read
    ) as p_read, patch.object(
        TestTransformer, "apply", side_effect=TestTransformer.apply
    ) as p_apply, patch.object(
        TestWriter, "write", side_effect=TestWriter.write
    ) as p_write:
        resp = client.post(
            "/vocabularies/tasks", headers=h, data=json.dumps(task_config)
        )
        assert resp.status_code == 202
        p_read.assert_called()
        p_apply.assert_called()
        p_write.assert_called()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Funders API tests."""

from copy import deepcopy
from functools import partial

import pytest
from invenio_search import current_search_client
from jsonschema import ValidationError

from invenio_vocabularies.contrib.funders.api import Funder


@pytest.fixture()
def search_get():
    """Get a document from an index."""
    return partial(current_search_client.get, Funder.index._name)


@pytest.fixture()
def example_funder(db, funder_full_data):
    """Example funder."""
    api_funder = deepcopy(funder_full_data)
    pid = api_funder.pop("id")  # at API level it's passed as an arg
    fun = Funder.create(api_funder, pid=pid)
    fun.commit()
    db.session.commit()
    return fun


def test_funder_schema_validation(app, example_funder):
    """Funder schema validation."""
    # valid data
    fun = example_funder

    assert fun.schema == "local://funders/funder-v1.0.0.json"
    assert fun.pid
    assert fun.id

    # invalid data
    examples = [
        # name must be a string
        {"name": 123},
        # country must be a string
        {"name": "cern", "country": 123},
    ]

    for ex in examples:
        pytest.raises(ValidationError, Funder.create, ex)


def test_funder_indexing(app, example_funder, indexer, search_get):
    """Test indexing of a funder."""
    # Index document in ES
    assert indexer.index(example_funder)["result"] == "created"

    # Retrieve document from ES
    data = search_get(id=example_funder.id)

    # Loads the ES data and compare
    fun = Funder.loads(data["_source"])
    assert fun == example_funder
    assert fun.id == example_funder.id
    assert fun.pid.pid_value == example_funder.pid.pid_value
    assert fun.revision_id == example_funder.revision_id
    assert fun.created == example_funder.created
    assert fun.updated == example_funder.updated


def test_funder_pid(app, example_funder):
    """Test funder pid creation."""

    assert example_funder.pid.pid_value == "01ggx4157"
    assert Funder.pid.resolve("01ggx4157") == example_funder



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2021 Northwestern University.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabularies configuration."""

import idutils
from invenio_i18n import lazy_gettext as _

from .datastreams.readers import (
    CSVReader,
    GzipReader,
    JsonLinesReader,
    JsonReader,
    TarReader,
    XMLReader,
    YamlReader,
    ZipReader,
)
from .datastreams.transformers import XMLTransformer
from .datastreams.writers import ServiceWriter, YamlWriter
from .resources.resource import VocabulariesResourceConfig
from .services.service import VocabulariesServiceConfig

VOCABULARIES_RESOURCE_CONFIG = VocabulariesResourceConfig
"""Configure the resource."""

VOCABULARIES_SERVICE_CONFIG = VocabulariesServiceConfig
"""Configure the service."""

VOCABULARIES_IDENTIFIER_SCHEMES = {
    "grid": {"label": _("GRID"), "validator": lambda x: True},
    "gnd": {"label": _("GND"), "validator": idutils.is_gnd},
    "isni": {"label": _("ISNI"), "validator": idutils.is_isni},
    "ror": {"label": _("ROR"), "validator": idutils.is_ror},
}
""""Generic identifier schemes, usable by other vocabularies."""

VOCABULARIES_AFFILIATION_SCHEMES = {
    **VOCABULARIES_IDENTIFIER_SCHEMES,
}
"""Affiliations allowed identifier schemes."""

VOCABULARIES_FUNDER_SCHEMES = {
    **VOCABULARIES_IDENTIFIER_SCHEMES,
    "doi": {"label": _("DOI"), "validator": idutils.is_doi},
}
"""Funders allowed identifier schemes."""

VOCABULARIES_FUNDER_DOI_PREFIX = "10.13039"
"""DOI prefix for the identifier formed with the FundRef id."""

VOCABULARIES_AWARD_SCHEMES = {
    "url": {"label": _("URL"), "validator": idutils.is_url},
    "doi": {"label": _("DOI"), "validator": idutils.is_doi},
}
"""Awards allowed identifier schemes."""

VOCABULARIES_AWARDS_OPENAIRE_FUNDERS = {
    "aka_________": "05k73zm37",
    "anr_________": "00rbzpz17",
    "arc_________": "05mmh0f86",
    "asap________": "03zj4c476",
    "cihr________": "01gavpb45",
    "corda_______": "00k4n6c32",
    "corda_____he": "00k4n6c32",
    "corda__h2020": "00k4n6c32",
    "euenvagency_": "02k4b9v70",
    "fct_________": "00snfqn58",
    "fwf_________": "013tf3c58",
    "inca________": "03m8vkq32",
    "irb_hr______": "03n51vw80",
    "mestd_______": "01znas443",
    "nhmrc_______": "011kf5r70",
    "nih_________": "01cwqze88",
    "nserc_______": "01h531d29",
    "nsf_________": "021nxhr62",
    "nwo_________": "04jsz6e67",
    "rcuk________": "00dq2kk65",  # deprecated funder org
    "sfi_________": "0271asj38",
    "snsf________": "00yjd3n13",
    "sshrc_______": "006cvnv84",
    "tubitakf____": "04w9kkr77",
    "ukri________": "001aqnf71",
    "wt__________": "029chgv08",
}
"""Mapping of OpenAIRE and ROR funder codes."""

VOCABULARIES_AWARDS_EC_ROR_ID = "00k4n6c32"
"""ROR ID for EC funder."""

VOCABULARIES_NAMES_SCHEMES = {
    "orcid": {"label": _("ORCID"), "validator": idutils.is_orcid, "datacite": "ORCID"},
    "isni": {"label": _("ISNI"), "validator": idutils.is_isni, "datacite": "ISNI"},
    "gnd": {"label": _("GND"), "validator": idutils.is_gnd, "datacite": "GND"},
}
"""Names allowed identifier schemes."""

VOCABULARIES_DATASTREAM_READERS = {
    "csv": CSVReader,
    "json": JsonReader,
    "jsonl": JsonLinesReader,
    "gzip": GzipReader,
    "tar": TarReader,
    "yaml": YamlReader,
    "zip": ZipReader,
    "xml": XMLReader,
}
"""Data Streams readers."""

VOCABULARIES_DATASTREAM_TRANSFORMERS = {
    "xml": XMLTransformer,
}
"""Data Streams transformers."""

VOCABULARIES_DATASTREAM_WRITERS = {
    "service": ServiceWriter,
    "yaml": YamlWriter,
}
"""Data Streams writers."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 Northwestern University.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Test the subject vocabulary service."""

import arrow
import pytest
from invenio_pidstore.errors import PIDDeletedError

from invenio_vocabularies.contrib.subjects.api import Subject


def test_subject_simple_flow(app, db, service, identity, subject_full_data):
    """Test a simple vocabulary service flow."""
    # Service
    assert service.id == "subjects"
    assert service.config.indexer_queue_name == "subjects"

    # Create a controlled subject
    item = service.create(identity, subject_full_data)
    id_ = item.id
    data = item.data

    assert id_ == subject_full_data["id"]
    for k, v in subject_full_data.items():
        assert data[k] == v, data

    # Read it
    read_item = service.read(identity, "https://id.nlm.nih.gov/mesh/D000001")
    assert item.id == read_item.id
    assert item.data == read_item.data

    # Refresh index to make changes live.
    Subject.index.refresh()

    # Search it
    # Because colons are special characters in a query, we must place
    # quotes around id_ which contains colons
    res = service.search(identity, q=f'id:"{id_}"', size=25, page=1)
    assert res.total == 1
    assert list(res.hits)[0] == read_item.data

    # Update it
    data = read_item.data
    data["subject"] = "Antibiotics"
    update_item = service.update(identity, id_, data)
    assert item.id == update_item.id
    assert update_item["subject"] == "Antibiotics"

    # Delete it
    assert service.delete(identity, id_)

    # Refresh to make changes live
    Subject.index.refresh()

    # Fail to retrieve it
    # - db
    pytest.raises(PIDDeletedError, service.read, identity, id_)
    # - search
    res = service.search(identity, q=f'id:"{id_}"', size=25, page=1)
    assert res.total == 0


def test_indexed_at_query(app, db, service, identity, subject_full_data):
    before = arrow.utcnow().strftime("%Y-%m-%dT%H:%M:%S.%f")
    _ = service.create(identity, subject_full_data)
    now = arrow.utcnow().strftime("%Y-%m-%dT%H:%M:%S.%f")
    Subject.index.refresh()

    # there is previous to before
    res = service.search(identity, q=f"indexed_at:[* TO {before}]", size=25, page=1)
    assert res.total == 0

    # there is previous to now
    res = service.search(identity, q=f"indexed_at:[* TO {now}]", size=25, page=1)
    assert res.total == 1



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Invenio vocabularies views."""

from flask import Blueprint

blueprint = Blueprint("invenio_vocabularies_ext", __name__)


@blueprint.record_once
def init(state):
    """Init app."""
    app = state.app
    # Register services - cannot be done in extension because
    # Invenio-Records-Resources might not have been initialized.
    sregistry = app.extensions["invenio-records-resources"].registry
    ext = app.extensions["invenio-vocabularies"]
    sregistry.register(ext.affiliations_service, service_id="affiliations")
    sregistry.register(ext.awards_service, service_id="awards")
    sregistry.register(ext.funders_service, service_id="funders")
    sregistry.register(ext.names_service, service_id="names")
    sregistry.register(ext.subjects_service, service_id="subjects")
    sregistry.register(ext.service, service_id="vocabularies")
    # Register indexers
    iregistry = app.extensions["invenio-indexer"].registry
    iregistry.register(ext.affiliations_service.indexer, indexer_id="affiliations")
    iregistry.register(ext.awards_service.indexer, indexer_id="awards")
    iregistry.register(ext.funders_service.indexer, indexer_id="funders")
    iregistry.register(ext.names_service.indexer, indexer_id="names")
    iregistry.register(ext.subjects_service.indexer, indexer_id="subjects")
    iregistry.register(ext.service.indexer, indexer_id="vocabularies")


def create_blueprint_from_app(app):
    """Create app blueprint."""
    return app.extensions["invenio-vocabularies"].resource.as_blueprint()


def create_affiliations_blueprint_from_app(app):
    """Create app blueprint."""
    return app.extensions["invenio-vocabularies"].affiliations_resource.as_blueprint()


def create_awards_blueprint_from_app(app):
    """Create app blueprint."""
    return app.extensions["invenio-vocabularies"].awards_resource.as_blueprint()


def create_funders_blueprint_from_app(app):
    """Create app blueprint."""
    return app.extensions["invenio-vocabularies"].funders_resource.as_blueprint()


def create_names_blueprint_from_app(app):
    """Create app blueprint."""
    return app.extensions["invenio-vocabularies"].names_resource.as_blueprint()


def create_subjects_blueprint_from_app(app):
    """Create app blueprint."""
    return app.extensions["invenio-vocabularies"].subjects_resource.as_blueprint()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Data Streams transformers tests."""

import pytest

from invenio_vocabularies.datastreams import StreamEntry
from invenio_vocabularies.datastreams.errors import TransformerError
from invenio_vocabularies.datastreams.transformers import XMLTransformer


@pytest.fixture(scope="module")
def expected_from_xml():
    return {
        "field_one": "value",
        "multi_field": {"some": "value", "another": "value too"},
    }


def test_xml_transformer(expected_from_xml):
    bytes_xml_entry = StreamEntry(
        bytes(
            '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\n'
            "<record:record>\n"
            "    <field_one:field_one>value</field_one:field_one>\n"
            "    <multi_field:multi_field>\n"
            "        <some:some>value</some:some>\n"
            "        <another:another>value too</another:another>\n"
            "    </multi_field:multi_field>\n"
            "</record:record>\n",
            encoding="raw_unicode_escape",
        )
    )

    transformer = XMLTransformer()
    assert expected_from_xml == transformer.apply(bytes_xml_entry).entry


def test_bad_xml_transformer():
    bytes_xml_entry = StreamEntry(
        bytes(
            '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\n'
            "<field_one:field_one>value</field_one:field_one>\n"
            "<multi_field:multi_field>\n"
            "    <some:some>value</some:some>\n"
            "    <another:another>value too</another:another>\n"
            "</multi_field:multi_field>\n",
            encoding="raw_unicode_escape",
        )
    )

    transformer = XMLTransformer()
    with pytest.raises(TransformerError):
        transformer.apply(bytes_xml_entry)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Names schema."""

from functools import partial

from invenio_i18n import lazy_gettext as _
from marshmallow import ValidationError, fields, post_load, validates_schema
from marshmallow_utils.fields import IdentifierSet, SanitizedUnicode
from marshmallow_utils.schemas import IdentifierSchema

from ...services.schema import BaseVocabularySchema, ModePIDFieldVocabularyMixin
from ..affiliations.schema import AffiliationRelationSchema
from .config import names_schemes


class NameSchema(BaseVocabularySchema, ModePIDFieldVocabularyMixin):
    """Service schema for names.

    Note that this vocabulary is very different from the base vocabulary
    so it does not inherit from it.
    """

    name = SanitizedUnicode()
    given_name = SanitizedUnicode()
    family_name = SanitizedUnicode()
    identifiers = IdentifierSet(
        fields.Nested(
            partial(
                IdentifierSchema,
                # It is intended to allow org schemes to be sent as personal
                # and viceversa. This is a trade off learnt from running
                # Zenodo in production.
                allowed_schemes=names_schemes,
            )
        )
    )
    affiliations = fields.List(fields.Nested(AffiliationRelationSchema))

    @validates_schema
    def validate_names(self, data, **kwargs):
        """Validate names."""
        can_compose = data.get("family_name") and data.get("given_name")
        name = data.get("name")
        if not can_compose and not name:
            messages = [
                _(
                    "A name or the family name together with the given name must be present."
                )
            ]
            raise ValidationError({"family_name": messages})

    @post_load
    def calculate_name(self, data, **kwargs):
        """Update names for person.

        Fill name from given_name and family_name if person.
        """
        family_name = data.get("family_name")
        given_name = data.get("given_name")
        if family_name and given_name:
            data["name"] = f"{family_name}, {given_name}"

        return data



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Writers module."""

from abc import ABC, abstractmethod
from pathlib import Path

import yaml
from invenio_access.permissions import system_identity
from invenio_pidstore.errors import PIDAlreadyExists
from invenio_records.systemfields.relations.errors import InvalidRelationValue
from invenio_records_resources.proxies import current_service_registry
from marshmallow import ValidationError

from .datastreams import StreamEntry
from .errors import WriterError


class BaseWriter(ABC):
    """Base writer."""

    @abstractmethod
    def write(self, stream_entry, *args, **kwargs):
        """Writes the input stream entry to the target output.

        :returns: A StreamEntry. The result of writing the entry.
                  Raises WriterException in case of errors.

        """
        pass


class ServiceWriter(BaseWriter):
    """Writes the entries to an RDM instance using a Service object."""

    def __init__(self, service_or_name, *args, identity=None, update=False, **kwargs):
        """Constructor.

        :param service_or_name: a service instance or a key of the
                                service registry.
        :param identity: access identity.
        :param update: if True it will update records if they exist.
        """
        if isinstance(service_or_name, str):
            service_or_name = current_service_registry.get(service_or_name)

        self._service = service_or_name
        self._identity = identity or system_identity
        self._update = update

        super().__init__(*args, **kwargs)

    def _entry_id(self, entry):
        """Get the id from an entry."""
        return (entry["type"], entry["id"])

    def _resolve(self, id_):
        return self._service.read(self._identity, id_)

    def write(self, stream_entry, *args, **kwargs):
        """Writes the input entry using a given service."""
        entry = stream_entry.entry
        try:
            try:
                return StreamEntry(self._service.create(self._identity, entry))
            except PIDAlreadyExists:
                if not self._update:
                    raise WriterError([f"Vocabulary entry already exists: {entry}"])
                vocab_id = self._entry_id(entry)
                current = self._resolve(vocab_id)
                updated = dict(current.to_dict(), **entry)
                return StreamEntry(
                    self._service.update(self._identity, vocab_id, updated)
                )

        except ValidationError as err:
            raise WriterError([{"ValidationError": err.messages}])
        except InvalidRelationValue as err:
            # TODO: Check if we can get the error message easier
            raise WriterError([{"InvalidRelationValue": err.args[0]}])


class YamlWriter(BaseWriter):
    """Writes the entries to a YAML file."""

    def __init__(self, filepath, *args, **kwargs):
        """Constructor.

        :param filepath: path of the output file.
        """
        self._filepath = Path(filepath)

        super().__init__(*args, **kwargs)

    def write(self, stream_entry, *args, **kwargs):
        """Writes the input stream entry using a given service."""
        with open(self._filepath, "a") as file:
            # made into array for safer append
            # will always read array (good for reader)
            yaml.safe_dump([stream_entry.entry], file)

        return stream_entry



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Awards vocabulary."""

from .resources import AwardsResource, AwardsResourceConfig
from .services import AwardsService, AwardsServiceConfig

__all__ = [
    "AwardsResource",
    "AwardsResourceConfig",
    "AwardsService",
    "AwardsServiceConfig",
]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Funders datastreams, transformers, writers and readers."""

from idutils import normalize_ror
from invenio_access.permissions import system_identity
from invenio_i18n import lazy_gettext as _

from ...datastreams.errors import TransformerError
from ...datastreams.transformers import BaseTransformer
from ...datastreams.writers import ServiceWriter
from .config import funder_fundref_doi_prefix, funder_schemes


class FundersServiceWriter(ServiceWriter):
    """Funders service writer."""

    def __init__(self, *args, **kwargs):
        """Constructor."""
        service_or_name = kwargs.pop("service_or_name", "funders")
        super().__init__(service_or_name=service_or_name, *args, **kwargs)

    def _entry_id(self, entry):
        """Get the id from an entry."""
        return entry["id"]


class RORTransformer(BaseTransformer):
    """Transforms a JSON ROR record into a funders record."""

    def apply(self, stream_entry, **kwargs):
        """Applies the transformation to the stream entry."""
        record = stream_entry.entry
        funder = {}

        funder["id"] = normalize_ror(record.get("id"))
        if not funder["id"]:
            raise TransformerError(_("Id not found in ROR entry."))

        funder["name"] = record.get("name")
        if not funder["name"]:
            raise TransformerError(_("Name not found in ROR entry."))

        country_code = record.get("country", {}).get("country_code")
        if country_code:
            funder["country"] = country_code

        funder["title"] = {"en": funder["name"]}
        for label in record.get("labels", []):
            funder["title"][label["iso639"]] = label["label"]

        # The ROR is always listed in identifiers, expected by serialization
        funder["identifiers"] = [{"identifier": funder["id"], "scheme": "ror"}]
        valid_schemes = set(funder_schemes.keys())
        fund_ref = "fundref"
        valid_schemes.add(fund_ref)
        for scheme, identifier in record.get("external_ids", {}).items():
            scheme = scheme.lower()
            if scheme in valid_schemes:
                value = identifier.get("preferred") or identifier.get("all")[0]
                if scheme == fund_ref:
                    value = f"{funder_fundref_doi_prefix}/{value}"
                    scheme = "doi"

                funder["identifiers"].append(
                    {
                        "identifier": value,
                        "scheme": scheme,
                    }
                )

        stream_entry.entry = funder
        return stream_entry


VOCABULARIES_DATASTREAM_TRANSFORMERS = {
    "ror-funder": RORTransformer,
}
"""ROR Data Streams transformers."""


VOCABULARIES_DATASTREAM_WRITERS = {
    "funders-service": FundersServiceWriter,
}
"""Funders Data Streams transformers."""


DATASTREAM_CONFIG = {
    "readers": [
        {
            "type": "zip",
            "args": {
                "regex": ".json$",
            },
        },
        {"type": "json"},
    ],
    "transformers": [
        {"type": "ror-funder"},
    ],
    "writers": [
        {
            "type": "funders-service",
            "args": {
                "identity": system_identity,
            },
        }
    ],
}
"""Data Stream configuration.

An origin is required for the reader.
"""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Funders schema."""

from functools import partial

from invenio_i18n import lazy_gettext as _
from marshmallow import (
    ValidationError,
    fields,
    post_load,
    pre_dump,
    validate,
    validates_schema,
)
from marshmallow_utils.fields import IdentifierSet, SanitizedUnicode
from marshmallow_utils.schemas import IdentifierSchema

from ...services.schema import BaseVocabularySchema, ContribVocabularyRelationSchema
from .config import funder_schemes


class FunderRelationSchema(ContribVocabularyRelationSchema):
    """Funder schema."""

    ftf_name = "name"
    parent_field_name = "funder"
    name = SanitizedUnicode(
        validate=validate.Length(min=1, error=_("Name cannot be blank."))
    )


class FunderSchema(BaseVocabularySchema):
    """Service schema for funders."""

    name = SanitizedUnicode(
        required=True, validate=validate.Length(min=1, error=_("Name cannot be blank."))
    )
    country = SanitizedUnicode()
    identifiers = IdentifierSet(
        fields.Nested(
            partial(
                IdentifierSchema,
                allowed_schemes=funder_schemes,
                identifier_required=False,
            )
        )
    )

    id = SanitizedUnicode(
        validate=validate.Length(min=1, error=_("PID cannot be blank."))
    )

    @validates_schema
    def validate_id(self, data, **kwargs):
        """Validates ID."""
        is_create = "record" not in self.context
        if is_create and "id" not in data:
            raise ValidationError(_("Missing PID."), "id")
        if not is_create:
            data.pop("id", None)

    @post_load(pass_many=False)
    def move_id(self, data, **kwargs):
        """Moves id to pid."""
        if "id" in data:
            data["pid"] = data.pop("id")
        return data

    @pre_dump(pass_many=False)
    def extract_pid_value(self, data, **kwargs):
        """Extracts the PID value."""
        data["id"] = data.pid.pid_value
        return data



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020-2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Test the vocabulary service."""

import pytest
from flask_principal import Identity, PermissionDenied
from invenio_access.permissions import any_user, system_identity

from invenio_vocabularies.records.api import Vocabulary


#
# Fixtures
#
@pytest.fixture()
def anyuser_idty():
    """Simple identity to interact with the service."""
    identity = Identity(1)
    identity.provides.add(any_user)
    return identity


#
# Tests
#
def test_permissions_readonly(anyuser_idty, lang_type, lang_data, service):
    """Read-only permission test.

    These tests will eventually need to be rewritten once a more advanced
    permission policy is put in place. Currently the permission policy only
    allows a system process to create/update/delete vocabularies.
    """
    # Create - only system allowed
    pytest.raises(PermissionDenied, service.create, anyuser_idty, lang_data)
    item = service.create(system_identity, lang_data)
    id_ = item.id

    # Read - both allowed
    item = service.read(anyuser_idty, ("languages", id_))
    item = service.read(system_identity, ("languages", id_))

    # Refresh index to make changes live.
    Vocabulary.index.refresh()

    # Search - both allowed
    res = service.search(anyuser_idty, type="languages", q=f"id:{id_}", size=25, page=1)
    assert res.total == 1
    res = service.search(
        system_identity, type="languages", q=f"id:{id_}", size=25, page=1
    )
    assert res.total == 1

    # Update - only system allowed
    data = item.data
    data["title"]["en"] = "New title"
    with pytest.raises(PermissionDenied):
        service.update(anyuser_idty, ("languages", id_), data)
    service.update(system_identity, ("languages", id_), data)

    # Delete - only system allowed
    with pytest.raises(PermissionDenied):
        service.delete(anyuser_idty, ("languages", id_))
    service.delete(system_identity, ("languages", id_))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Celery tasks."""

from celery import shared_task
from flask import current_app

from ..datastreams.factories import DataStreamFactory


@shared_task(ignore_result=True)
def process_datastream(config):
    """Process a datastream from config."""
    ds = DataStreamFactory.create(
        readers_config=config["readers"],
        transformers_config=config.get("transformers"),
        writers_config=config["writers"],
    )

    for result in ds.process():
        if result.errors:
            for err in result.errors:
                current_app.logger.error(err)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Awards datastreams, transformers, writers and readers."""

from invenio_access.permissions import system_identity
from invenio_i18n import lazy_gettext as _

from ...datastreams.errors import TransformerError
from ...datastreams.transformers import BaseTransformer
from ...datastreams.writers import ServiceWriter
from .config import awards_ec_ror_id, awards_openaire_funders_mapping


class AwardsServiceWriter(ServiceWriter):
    """Funders service writer."""

    def __init__(self, *args, **kwargs):
        """Constructor."""
        service_or_name = kwargs.pop("service_or_name", "awards")
        super().__init__(service_or_name=service_or_name, *args, **kwargs)

    def _entry_id(self, entry):
        """Get the id from an entry."""
        return entry["id"]


class OpenAIREProjectTransformer(BaseTransformer):
    """Transforms an OpenAIRE project record into an award record."""

    def apply(self, stream_entry, **kwargs):
        """Applies the transformation to the stream entry."""
        record = stream_entry.entry
        award = {}

        code = record["code"]
        openaire_funder_prefix = record["id"].split("::")[0].split("|")[1]
        funder_id = awards_openaire_funders_mapping.get(openaire_funder_prefix)
        if funder_id is None:
            raise TransformerError(
                _(
                    "Unknown OpenAIRE funder prefix {openaire_funder_prefix}".format(
                        openaire_funder_prefix=openaire_funder_prefix
                    )
                )
            )

        award["id"] = f"{funder_id}::{code}"

        funding = next(iter(record.get("funding", [])), None)
        if funding:
            funding_stream_id = funding.get("funding_stream", {}).get("id", "")
            # Example funding stream ID: `EC::HE::HORIZON-AG-UN`. We need the `EC`
            # string, i.e. the second "part" of the identifier.
            program = next(iter(funding_stream_id.split("::")[1:2]), "")
            if program:
                award["program"] = program

        identifiers = []
        if funder_id == awards_ec_ror_id:
            identifiers.append(
                {
                    "identifier": f"https://cordis.europa.eu/projects/{code}",
                    "scheme": "url",
                }
            )
        elif record.get("websiteurl"):
            identifiers.append(
                {"identifier": record.get("websiteurl"), "scheme": "url"}
            )

        if identifiers:
            award["identifiers"] = identifiers

        award["number"] = code
        award["title"] = {"en": record["title"]}
        award["funder"] = {"id": funder_id}
        acronym = record.get("acronym")
        if acronym:
            award["acronym"] = acronym

        stream_entry.entry = award
        return stream_entry


VOCABULARIES_DATASTREAM_TRANSFORMERS = {
    "openaire-award": OpenAIREProjectTransformer,
}
"""ORCiD Data Streams transformers."""

VOCABULARIES_DATASTREAM_WRITERS = {
    "awards-service": AwardsServiceWriter,
}
"""ORCiD Data Streams transformers."""

DATASTREAM_CONFIG = {
    "readers": [
        {
            "type": "tar",
            "args": {
                "regex": ".json.gz$",
                "mode": "r",
            },
        },
        {"type": "gzip"},
        {"type": "jsonl"},
    ],
    "transformers": [
        {"type": "openaire-award"},
    ],
    "writers": [
        {
            "type": "awards-service",
            "args": {
                "identity": system_identity,
            },
        }
    ],
}
"""Data Stream configuration.

An origin is required for the reader.
"""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Base data stream."""

from .errors import ReaderError, TransformerError, WriterError


class StreamEntry:
    """Object to encapsulate streams processing."""

    def __init__(self, entry, errors=None):
        """Constructor."""
        self.entry = entry
        self.filtered = False
        self.errors = errors or []


class DataStream:
    """Data stream."""

    def __init__(self, readers, writers, transformers=None, *args, **kwargs):
        """Constructor.

        :param readers: an ordered list of readers.
        :param writers: an ordered list of writers.
        :param transformers: an ordered list of transformers to apply.
        """
        self._readers = readers
        self._transformers = transformers
        self._writers = writers

    def filter(self, stream_entry, *args, **kwargs):
        """Checks if an stream_entry should be filtered out (skipped)."""
        return False

    def process(self, *args, **kwargs):
        """Iterates over the entries.

        Uses the reader to get the raw entries and transforms them.
        It will iterate over the `StreamEntry` objects returned by
        the reader, apply the transformations and yield the result of
        writing it.
        """
        for stream_entry in self.read():
            if stream_entry.errors:
                yield stream_entry  # reading errors
            else:
                transformed_entry = self.transform(stream_entry)
                if transformed_entry.errors:
                    yield transformed_entry
                elif self.filter(transformed_entry):
                    transformed_entry.filtered = True
                    yield transformed_entry
                else:
                    yield self.write(transformed_entry)

    def read(self):
        """Recursively read the entries."""

        def pipe_gen(gen_funcs, piped_item=None):
            _gen_funcs = list(gen_funcs)  # copy to avoid modifying ref list
            # use and remove the current generator
            current_gen_func = _gen_funcs.pop(0)
            for item in current_gen_func(piped_item):
                try:
                    # exhaust iterations of subsequent generators
                    if _gen_funcs:
                        yield from pipe_gen(_gen_funcs, piped_item=item)
                    # there is no subsequent generator, return the current item
                    else:
                        yield StreamEntry(item)
                except ReaderError as err:
                    yield StreamEntry(
                        entry=item,
                        errors=[f"{current_gen_func.__qualname__}: {str(err)}"],
                    )

        read_gens = [r.read for r in self._readers]
        yield from pipe_gen(read_gens)

    def transform(self, stream_entry, *args, **kwargs):
        """Apply the transformations to an stream_entry."""
        for transformer in self._transformers:
            try:
                stream_entry = transformer.apply(stream_entry)
            except TransformerError as err:
                stream_entry.errors.append(
                    f"{transformer.__class__.__name__}: {str(err)}"
                )
                return stream_entry  # break loop

        return stream_entry

    def write(self, stream_entry, *args, **kwargs):
        """Apply the transformations to an stream_entry."""
        for writer in self._writers:
            try:
                writer.write(stream_entry)
            except WriterError as err:
                stream_entry.errors.append(f"{writer.__class__.__name__}: {str(err)}")

        return stream_entry

    def total(self, *args, **kwargs):
        """The total of entries obtained from the origin."""
        raise NotImplementedError()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Service schemas tests."""

import pytest
from marshmallow import ValidationError

from invenio_vocabularies.services.schema import TaskSchema


def test_valid_minimal():
    valid_minimal = {"readers": [{"type": "test"}], "writers": [{"type": "test"}]}
    assert valid_minimal == TaskSchema().load(valid_minimal)


def test_valid_full():
    valid_full = {
        "readers": [{"type": "test", "args": {"one": 1, "two": "two"}}],
        "transformers": [
            {"type": "test", "args": {"one": 1, "two": "two"}},
            {"type": "testtoo", "args": {"one": "1", "two": 2}},
        ],
        "writers": [
            {"type": "test", "args": {"one": 1, "two": "two"}},
            {"type": "testtoo", "args": {"one": "1", "two": 2}},
        ],
    }
    assert valid_full == TaskSchema().load(valid_full)


@pytest.mark.parametrize(
    "invalid",
    [
        {},
        {
            "readers": {"type": "test"},  # non-list readers
            "writers": [{"type": "test"}],
        },
        {
            "readers": [{"type": "testtoo"}],
            "writers": {"type": "test"},  # non-list writers
        },
        {
            "readers": [{"type": "testtoo"}],
            "transformers": {"type": "test"},  # non-list transformers
            "writers": [{"type": "test"}],
        },
        {
            "readers": [{"type": "testtoo"}],  # no writers
        },
        {
            "writers": [{"type": "test"}],  # no readers
        },
    ],
)
def test_invalid_empty(invalid):
    with pytest.raises(ValidationError):
        data = TaskSchema().load(invalid)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 Northwestern University.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Subjects schema tests."""

from copy import copy

import pytest
from marshmallow import ValidationError

from invenio_vocabularies.contrib.subjects.schema import (
    SubjectRelationSchema,
    SubjectSchema,
)


class TestSubjectSchema:
    def test_valid_full(self, subject_full_data):
        loaded = SubjectSchema().load(subject_full_data)
        assert subject_full_data == loaded

    def test_invalid_missing_field(self, subject_full_data):
        # no id
        invalid = copy(subject_full_data)
        del invalid["id"]
        with pytest.raises(ValidationError):
            SubjectSchema().load(invalid)

        # no scheme
        invalid = copy(subject_full_data)
        del invalid["scheme"]
        with pytest.raises(ValidationError):
            SubjectSchema().load(invalid)

        # no subject
        invalid = copy(subject_full_data)
        del invalid["subject"]
        with pytest.raises(ValidationError):
            SubjectSchema().load(invalid)


class TestSubjectRelationSchema:
    def test_valid_id(self):
        valid_id = {
            "id": "test",
        }
        assert valid_id == SubjectRelationSchema().load(valid_id)

    def test_valid_subject(self):
        valid_subject = {"subject": "Entity One"}
        assert valid_subject == SubjectRelationSchema().load(valid_subject)

    def test_invalid_empty(self):
        invalid_empty = {}
        with pytest.raises(ValidationError):
            SubjectRelationSchema().load(invalid_empty)

    def test_dump(self):
        subject = {"subject": "Entity One"}
        assert {"subject": "Entity One"} == SubjectRelationSchema().dump(subject)

        subject = {"subject": "Entity One", "scheme": "scheme A"}
        assert subject == SubjectRelationSchema().dump(subject)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Invenio module for managing vocabularies."""

from . import config
from .contrib.affiliations import (
    AffiliationsResource,
    AffiliationsResourceConfig,
    AffiliationsService,
    AffiliationsServiceConfig,
)
from .contrib.awards import (
    AwardsResource,
    AwardsResourceConfig,
    AwardsService,
    AwardsServiceConfig,
)
from .contrib.funders import (
    FundersResource,
    FundersResourceConfig,
    FundersService,
    FundersServiceConfig,
)
from .contrib.names import (
    NamesResource,
    NamesResourceConfig,
    NamesService,
    NamesServiceConfig,
)
from .contrib.subjects import (
    SubjectsResource,
    SubjectsResourceConfig,
    SubjectsService,
    SubjectsServiceConfig,
)
from .resources.resource import VocabulariesResource
from .services.service import VocabulariesService


class InvenioVocabularies(object):
    """Invenio-Vocabularies extension."""

    def __init__(self, app=None):
        """Extension initialization."""
        self.resource = None
        self.service = None
        if app:
            self.init_app(app)

    def init_app(self, app):
        """Flask application initialization."""
        self.init_config(app)
        self.init_services(app)
        self.init_resource(app)
        app.extensions["invenio-vocabularies"] = self

    def init_config(self, app):
        """Initialize configuration."""
        for k in dir(config):
            if k.startswith("VOCABULARIES_"):
                app.config.setdefault(k, getattr(config, k))

    def service_configs(self, app):
        """Customized service configs."""

        class ServiceConfigs:
            affiliations = AffiliationsServiceConfig
            awards = AwardsServiceConfig
            funders = FundersServiceConfig
            names = NamesServiceConfig
            subjects = SubjectsServiceConfig

        return ServiceConfigs

    def init_services(self, app):
        """Initialize vocabulary resources."""
        service_configs = self.service_configs(app)

        # Services
        self.affiliations_service = AffiliationsService(
            config=service_configs.affiliations,
        )
        self.awards_service = AwardsService(
            config=service_configs.awards,
        )
        self.funders_service = FundersService(config=service_configs.funders)
        self.names_service = NamesService(config=service_configs.names)
        self.subjects_service = SubjectsService(config=service_configs.subjects)
        self.service = VocabulariesService(
            config=app.config["VOCABULARIES_SERVICE_CONFIG"],
        )

    def init_resource(self, app):
        """Initialize vocabulary resources."""
        # Generic Vocabularies
        self.affiliations_resource = AffiliationsResource(
            service=self.affiliations_service,
            config=AffiliationsResourceConfig,
        )
        self.funders_resource = FundersResource(
            service=self.funders_service,
            config=FundersResourceConfig,
        )
        self.names_resource = NamesResource(
            service=self.names_service,
            config=NamesResourceConfig,
        )
        self.awards_resource = AwardsResource(
            service=self.awards_service,
            config=AwardsResourceConfig,
        )
        self.subjects_resource = SubjectsResource(
            service=self.subjects_service,
            config=SubjectsResourceConfig,
        )
        self.resource = VocabulariesResource(
            service=self.service,
            config=app.config["VOCABULARIES_RESOURCE_CONFIG"],
        )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2021 TU Wien.
# Copyright (C) 2021 Northwestern University.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test alembic recipes for Invenio-Vocabularies."""

import pytest
from invenio_db.utils import drop_alembic_version_table


def assert_alembic(alembic, table_excludes, db):
    """Assert that metadata of alembic and db matches.

    This method allows omitting tables dynamically created for tests.
    """
    # We exclude assert checks in case we test mysql backend as there is an error
    # with mock_metadata table not being created i.e results to a "NoSuchTable()"
    # error when `alembic.compare_metadata()` runs
    if is_mysql_engine(db):
        return
    assert not list(
        filter(
            # x[0] is the operation and x[1] is the table
            lambda x: x[0] == "add_table" and x[1].name not in table_excludes,
            alembic.compare_metadata(),
        )
    )


def is_mysql_engine(db):
    """Helper function to return if mysql engine is the db backend."""
    return db.engine.name == "mysql"


def test_alembic(app, database):
    """Test alembic recipes."""
    db = database
    ext = app.extensions["invenio-db"]

    if db.engine.name == "sqlite":
        raise pytest.skip("Upgrades are not supported on SQLite.")

    # Check that this package's SQLAlchemy models have been properly registered
    tables = [x.name for x in db.get_tables_for_bind()]
    assert "vocabularies_metadata" in tables
    assert "vocabularies_types" in tables
    assert "vocabularies_schemes" in tables

    # Specific vocabularies models
    assert "subject_metadata" in tables
    assert "affiliation_metadata" in tables
    assert "name_metadata" in tables
    assert "funder_metadata" in tables
    assert "award_metadata" in tables

    # Check that Alembic agrees that there's no further tables to create.
    assert_alembic(ext.alembic, ["mock_metadata"], db)

    # Drop everything and recreate tables all with Alembic
    db.drop_all()
    drop_alembic_version_table()
    ext.alembic.upgrade()
    assert_alembic(ext.alembic, ["mock_metadata"], db)

    # Try to upgrade and downgrade
    ext.alembic.stamp()
    ext.alembic.downgrade(target="96e796392533")
    ext.alembic.upgrade()
    assert_alembic(ext.alembic, ["mock_metadata"], db)

    # Cleanup
    drop_alembic_version_table()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2021 Northwestern University.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary service."""

from invenio_cache import current_cache
from invenio_db import db
from invenio_i18n import lazy_gettext as _
from invenio_records_resources.services import (
    Link,
    LinksTemplate,
    RecordService,
    RecordServiceConfig,
    SearchOptions,
    pagination_links,
)
from invenio_records_resources.services.records.components import DataComponent
from invenio_records_resources.services.records.params import (
    FilterParam,
    SuggestQueryParser,
)
from invenio_records_resources.services.records.schema import ServiceSchemaWrapper
from invenio_records_resources.services.uow import unit_of_work
from invenio_search.engine import dsl

from ..records.api import Vocabulary
from ..records.models import VocabularyType
from .components import PIDComponent, VocabularyTypeComponent
from .permissions import PermissionPolicy
from .schema import TaskSchema, VocabularySchema
from .tasks import process_datastream


class VocabularySearchOptions(SearchOptions):
    """Search options."""

    params_interpreters_cls = [
        FilterParam.factory(param="tags", field="tags"),
    ] + SearchOptions.params_interpreters_cls

    suggest_parser_cls = SuggestQueryParser.factory(
        fields=[
            "id.text^100",
            "id.text._2gram",
            "id.text._3gram",
            "title.en^5",
            "title.en._2gram",
            "title.en._3gram",
        ],
    )

    sort_default = "bestmatch"

    sort_default_no_query = "title"

    sort_options = {
        "bestmatch": dict(
            title=_("Best match"),
            fields=["_score"],  # ES defaults to desc on `_score` field
        ),
        "title": dict(
            title=_("Title"),
            fields=["title_sort"],
        ),
        "newest": dict(
            title=_("Newest"),
            fields=["-created"],
        ),
        "oldest": dict(
            title=_("Oldest"),
            fields=["created"],
        ),
    }


class VocabulariesServiceConfig(RecordServiceConfig):
    """Vocabulary service configuration."""

    service_id = "vocabularies"
    indexer_queue_name = "vocabularies"
    permission_policy_cls = PermissionPolicy
    record_cls = Vocabulary
    schema = VocabularySchema
    task_schema = TaskSchema

    search = VocabularySearchOptions

    components = [
        # Order of components are important!
        VocabularyTypeComponent,
        DataComponent,
        PIDComponent,
    ]

    links_item = {
        "self": Link(
            "{+api}/vocabularies/{type}/{id}",
            vars=lambda record, vars: vars.update(
                {
                    "id": record.pid.pid_value,
                    "type": record.type.id,
                }
            ),
        ),
    }

    links_search = pagination_links("{+api}/vocabularies/{type}{?args*}")


class VocabulariesService(RecordService):
    """Vocabulary service."""

    @property
    def task_schema(self):
        """Returns the data schema instance."""
        return ServiceSchemaWrapper(self, schema=self.config.task_schema)

    @unit_of_work()
    def create_type(self, identity, id, pid_type, uow=None):
        """Create a new vocabulary type."""
        self.require_permission(identity, "manage")
        type_ = VocabularyType.create(id=id, pid_type=pid_type)
        return type_

    def search(
        self, identity, params=None, search_preference=None, type=None, **kwargs
    ):
        """Search for vocabulary entries."""
        self.require_permission(identity, "search")

        # If not found, NoResultFound is raised (caught by the resource).
        vocabulary_type = VocabularyType.query.filter_by(id=type).one()

        # Prepare and execute the search
        params = params or {}
        search_result = self._search(
            "search",
            identity,
            params,
            search_preference,
            extra_filter=dsl.Q("term", type__id=vocabulary_type.id),
            **kwargs
        ).execute()

        return self.result_list(
            self,
            identity,
            search_result,
            params,
            links_tpl=LinksTemplate(
                self.config.links_search,
                context={
                    "args": params,
                    "type": vocabulary_type.id,
                },
            ),
            links_item_tpl=self.links_item_tpl,
        )

    def read_all(self, identity, fields, type, cache=True, extra_filter="", **kwargs):
        """Search for records matching the querystring."""
        cache_key = type + "_" + str(extra_filter) + "_" + "-".join(fields)
        results = current_cache.get(cache_key)
        search_query = dsl.Q("match_all")

        if not results:
            # If not found, NoResultFound is raised (caught by the resource).
            vocabulary_type = VocabularyType.query.filter_by(id=type).one()
            vocab_id_filter = dsl.Q("term", type__id=vocabulary_type.id)
            if extra_filter:
                vocab_id_filter = vocab_id_filter & extra_filter
            results = self._read_many(
                identity, search_query, fields, extra_filter=vocab_id_filter, **kwargs
            )
            if cache:
                # ES DSL Response is not pickable.
                # If saved in cache serialization wont work with to_dict()
                current_cache.set(cache_key, results.to_dict())

        else:
            search = self.create_search(
                identity=identity,
                record_cls=self.record_cls,
                search_opts=self.config.search,
                permission_action="search",
            ).query(search_query)

            results = dsl.response.Response(search, results)

        return self.result_list(self, identity, results)

    def read_many(self, identity, type, ids, fields=None, **kwargs):
        """Search for records matching the querystring filtered by ids."""
        search_query = dsl.Q("match_all")
        vocabulary_type = VocabularyType.query.filter_by(id=type).one()
        vocab_id_filter = dsl.Q("term", type__id=vocabulary_type.id)
        filters = []

        for id_ in ids:
            filters.append(dsl.Q("term", **{"id": id_}))
        filter = dsl.Q("bool", minimum_should_match=1, should=filters)
        filter = filter & vocab_id_filter
        results = self._read_many(
            identity, search_query, fields, extra_filter=filter, **kwargs
        )

        return self.result_list(self, identity, results)

    def launch(self, identity, data):
        """Create a task.

        FIXME: This is a PoC. The final implementation should resemble
        the PIDs, having a sub-service and a manager. If persistance is
        added UoW should be used.
        """
        self.require_permission(identity, "manage")
        task_config, _ = self.task_schema.load(
            data,
            context={"identity": identity},  # FIXME: is this needed
            raise_errors=True,
        )
        process_datastream.delay(task_config)

        # 202 if accepted, otherwise it will be caught by an error handler
        return True



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
# Copyright (C) 2022 Northwestern University.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Affiliations search mappings."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""CLI Module tests."""

import tarfile
from pathlib import Path

import pytest
from invenio_access.permissions import system_identity
from invenio_records_resources.proxies import current_service_registry

from invenio_vocabularies.cli import _process_vocab, get_config_for_ds, vocabularies
from invenio_vocabularies.config import (
    VOCABULARIES_DATASTREAM_TRANSFORMERS,
    VOCABULARIES_DATASTREAM_WRITERS,
)
from invenio_vocabularies.contrib.names.api import Name
from invenio_vocabularies.contrib.names.datastreams import (
    VOCABULARIES_DATASTREAM_TRANSFORMERS as NAMES_TRANSFORMERS,
)
from invenio_vocabularies.contrib.names.datastreams import (
    VOCABULARIES_DATASTREAM_WRITERS as NAMES_WRITERS,
)


@pytest.fixture(scope="module")
def app_config(app_config):
    """Mimic an instance's configuration."""
    app_config["VOCABULARIES_DATASTREAM_TRANSFORMERS"] = {
        **VOCABULARIES_DATASTREAM_TRANSFORMERS,
        **NAMES_TRANSFORMERS,
    }
    app_config["VOCABULARIES_DATASTREAM_WRITERS"] = {
        **VOCABULARIES_DATASTREAM_WRITERS,
        **NAMES_WRITERS,
    }

    return app_config


@pytest.fixture(scope="module")
def name_xml():
    return (
        '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\n'
        '<record:record path="/0000-0001-8135-3489">\n'
        "    <common:orcid-identifier>\n"
        "        <common:uri>https://orcid.org/0000-0001-8135-3489</common:uri>\n"  # noqa
        "        <common:path>0000-0001-8135-3489</common:path>\n"
        "        <common:host>orcid.org</common:host>\n"
        "    </common:orcid-identifier>\n"
        '    <person:person path="/0000-0001-8135-3489/person">\n'
        '        <person:name visibility="public" path="0000-0001-8135-3489">\n'  # noqa
        "            <personal-details:given-names>Lars Holm</personal-details:given-names>"  # noqa
        "            <personal-details:family-name>Nielsen</personal-details:family-name>\n"  # noqa
        "        </person:name>\n"
        '        <external-identifier:external-identifiers path="/0000-0001-8135-3489/external-identifiers"/>\n'  # noqa
        "    </person:person>\n"
        '    <activities:activities-summary path="/0000-0001-8135-3489/activities">\n'  # noqa
        '       <activities:employments path="/0000-0001-8135-3489/employments">\n'  # noqa
        "           <employments:affiliation-group>\n"
        "               <employments:employment-summary>\n"
        "                   <employment-summary:organization>\n"
        "                       <organization:name>CERN</organization:name>\n"
        "                   </employment-summary:organization>\n"
        "               </employments:employment-summary>\n"
        "           </employments:affiliation-group>\n"
        "       </activities:employments>\n"
        "    </activities:activities-summary>\n"
        "</record:record>\n"
    )


@pytest.fixture(scope="function")
def names_tar_file(name_xml):
    """Creates a Tar file with three files (two yaml) inside.

    Each iteration should return the content of one yaml file,
    it should ignore the .other file.
    """
    filename = Path("cli_test.tar.gz")
    with tarfile.open(filename, "w:gz") as tar:
        inner_filename = Path("lnielsen_name.xml")
        with open(inner_filename, "w") as file:
            file.write(name_xml)
        tar.add(inner_filename)
        inner_filename.unlink()

    yield filename

    filename.unlink()  # delete created file


def test_process(app, names_tar_file):
    service = current_service_registry.get("names")
    config = get_config_for_ds(vocabulary="names", origin=names_tar_file.absolute())
    _process_vocab(config)
    Name.index.refresh()

    orcid = "0000-0001-8135-3489"
    results = service.search(system_identity, q=f"identifiers.identifier:{orcid}")

    assert results.total == 1
    assert list(results.hits)[0]["identifiers"][0]["identifier"] == orcid


def test_update_cmd(app, names_tar_file):
    # cli update
    runner = app.test_cli_runner()
    result = runner.invoke(
        vocabularies,
        ["update", "-v", "names", "--origin", names_tar_file.absolute()],
    )
    assert result.exit_code == 0



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Names Marshmallow schema tests."""

import pytest
from marshmallow import ValidationError

from invenio_vocabularies.contrib.names.schema import NameSchema


def test_valid_full(app, name_full_data):
    loaded = NameSchema().load(name_full_data)
    name_full_data["pid"] = name_full_data.pop("id")
    assert name_full_data == loaded


def test_valid_minimal(app):
    data = {
        "id": "0000-0001-8135-3489",
        "name": "Doe, John",
    }
    loaded = NameSchema().load(data)
    assert {
        "pid": "0000-0001-8135-3489",
        "name": "Doe, John",
    } == loaded

    data = {
        "id": "0000-0001-8135-3489",
        "family_name": "Doe",
        "given_name": "John",
    }
    loaded = NameSchema().load(data)
    assert {
        "pid": "0000-0001-8135-3489",
        "family_name": "Doe",
        "given_name": "John",
        "name": "Doe, John",
    } == loaded


def test_invalid_no_names(app):
    # no name
    invalid = {
        "identifiers": [{"identifier": "0000-0001-8135-3489", "scheme": "orcid"}],
        "affiliations": [{"id": "cern"}, {"name": "CustomORG"}],
    }
    with pytest.raises(ValidationError):
        data = NameSchema().load(invalid)

    # only given name
    invalid["given_name"] = ("John",)

    with pytest.raises(ValidationError):
        data = NameSchema().load(invalid)

    # only family name
    invalid["family_name"] = ("Doe",)
    invalid.pop("given_name")

    with pytest.raises(ValidationError):
        data = NameSchema().load(invalid)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""REST API search tests."""

import pytest

from invenio_vocabularies.records.api import Vocabulary


#
# Fixtures
#
@pytest.fixture(scope="module")
def example_data():
    """Example data for records."""
    return [
        {"id": "text", "title": {"en": "Text"}, "type": "resourcetypes"},
        {
            "id": "data",
            "title": {"en": "Data"},
            "type": "resourcetypes",
            "tags": ["recommended"],
        },
    ]


@pytest.fixture(scope="module")
def example_records(database, identity, service, example_data):
    """Create some example records."""
    service.create_type(identity, "resourcetypes", "rt")
    records = []
    for data in example_data:
        records.append(service.create(identity, data))
    Vocabulary.index.refresh()
    return records


@pytest.fixture()
def prefix():
    """API prefix."""
    return "/vocabularies/resourcetypes"


#
# Tests
#
def test_invalid_type(app, client, h):
    """Test invalid type."""
    # invalid type
    res = client.get("/vocabularies/invalid", headers=h)
    assert res.status_code == 404
    # trailing slash
    res = client.get("/vocabularies/invalid/", headers=h)
    assert res.status_code == 404


def test_type_as_query_arg(app, lang_type, example_records, client, h, prefix):
    """Test passing an invalid query parameter."""
    # Ensure we cannot pass invalid query parameters
    res = client.get(f"{prefix}?invalid=test", headers=h)
    assert res.status_code == 200
    assert "invalid" not in res.json["links"]["self"]

    # It should not be possible to pass 'type' in query args (because it's in
    # the URL route instead)
    res = client.get(f"{prefix}?type={lang_type.id}", headers=h)
    assert res.status_code == 200
    assert lang_type.id not in res.json["links"]["self"]
    assert "type=" not in res.json["links"]["self"]
    assert "resourcetypes" in res.json["links"]["self"]


def test_search(client, example_records, h, prefix):
    """Test a successful search."""
    res = client.get(prefix, headers=h)
    assert res.status_code == 200
    assert res.json["hits"]["total"] == 2
    assert res.json["sortBy"] == "title"


def test_query_q(client, example_records, h, prefix):
    """Test a successful search."""
    # Test query (q=)
    res = client.get(f"{prefix}?q=title.en:text", headers=h)
    assert res.status_code == 200
    assert res.json["hits"]["total"] == 1
    assert res.json["sortBy"] == "bestmatch"

    # Test sort
    res = client.get(f"{prefix}?q=*&sort=bestmatch", headers=h)
    assert res.status_code == 200
    assert res.json["sortBy"] == "bestmatch"

    # Test size
    res = client.get(f"{prefix}?size=1&page=1", headers=h)
    assert res.status_code == 200
    assert res.json["hits"]["total"] == 2
    assert len(res.json["hits"]["hits"]) == 1
    assert "next" in res.json["links"]


def test_tags_filter(client, example_records, h, prefix):
    """Test filter on tags."""
    # Test query (q=)
    res = client.get(f"{prefix}?tags=recommended", headers=h)
    assert res.status_code == 200
    assert res.json["hits"]["total"] == 1



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Test the vocabulary service facets."""

import pytest

from invenio_vocabularies.records.api import Vocabulary
from invenio_vocabularies.services.facets import VocabularyLabels


@pytest.fixture(scope="module")
def example_data():
    """Example data."""
    return [
        {
            "id": "eng",
            "title": {"en": "English", "da": "Engelsk"},
            "type": "languages",
        },
        {
            "id": "spa",
            "title": {"en": "Spanish", "da": "Spansk"},
            "type": "languages",
        },
        {
            "id": "fra",
            "title": {"en": "French", "da": "Fransk"},
            "type": "languages",
        },
    ]


@pytest.fixture(scope="module")
def example_records(database, identity, service, example_data):
    """Create some example records."""
    service.create_type(identity, "languages", "lng")
    records = []
    for data in example_data:
        records.append(service.create(identity, data))

    Vocabulary.index.refresh()
    return records


def _assert_vocabs(labels_instance):
    labels = labels_instance(["eng", "spa"])

    assert labels["eng"] == "English"
    assert labels["spa"] == "Spanish"


def test_vocabulary_label_cache(example_records):
    # Uses example record to have two languages created on the vocabulary
    labels_instance = VocabularyLabels(vocabulary="languages", cache=True)
    _assert_vocabs(labels_instance)


def test_vocabulary_label_no_cache(example_records):
    # Uses example record to have two languages created on the vocabulary
    labels_instance = VocabularyLabels(vocabulary="languages", cache=False)
    _assert_vocabs(labels_instance)


def test_vocabulary_label_not_found(example_records):
    # Uses example record to have two languages created on the vocabulary
    labels_instance = VocabularyLabels(vocabulary="languages", cache=True)
    labels = labels_instance(["eng", "glg"])

    assert labels["eng"] == "English"
    assert not labels.get("glg")  # search wont fail



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary awards facets and labels."""

from ...services.facets import VocabularyLabels, lazy_get_label


class FundersLabels(VocabularyLabels):
    """Fetching of vocabulary labels for facets."""

    def __init__(self, vocabulary, cache=False, service_name=None):
        """Initialize the labels."""
        super().__init__(
            vocabulary,
            cache,
            service_name="funders",
        )
        self.fields = ["id", "title", "country"]  # not configurable

    def _vocab_to_label(self, vocab):
        """Returns the label string for a vocabulary entry."""
        return f"{lazy_get_label(vocab['title'])} ({vocab['country']})"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Names datastreams, transformers, writers and readers."""

from invenio_access.permissions import system_identity
from invenio_records.dictutils import dict_lookup

from ...datastreams.errors import TransformerError
from ...datastreams.readers import SimpleHTTPReader
from ...datastreams.transformers import BaseTransformer
from ...datastreams.writers import ServiceWriter


class OrcidHTTPReader(SimpleHTTPReader):
    """ORCiD HTTP Reader."""

    def __init__(self, *args, test_mode=True, **kwargs):
        """Constructor."""
        if test_mode:
            origin = "https://sandbox.orcid.org/{id}"
        else:
            origin = "https://orcid.org/{id}"

        super().__init__(origin, *args, **kwargs)


class OrcidTransformer(BaseTransformer):
    """Transforms an ORCiD record into a names record."""

    def apply(self, stream_entry, **kwargs):
        """Applies the transformation to the stream entry."""
        record = stream_entry.entry
        person = record["person"]
        orcid_id = record["orcid-identifier"]["path"]

        name = person.get("name")
        if name is None:
            raise TransformerError(f"Name not found in ORCiD entry.")

        entry = {
            "id": orcid_id,
            "given_name": name.get("given-names"),
            "family_name": name.get("family-name"),
            "identifiers": [{"scheme": "orcid", "identifier": orcid_id}],
            "affiliations": [],
        }

        try:
            employments = dict_lookup(
                record, "activities-summary.employments.affiliation-group"
            )
            if isinstance(employments, dict):
                employments = [employments]
            history = set()
            for employment in employments:
                terminated = employment["employment-summary"].get("end-date")
                affiliation = dict_lookup(
                    employment,
                    "employment-summary.organization.name",
                )
                if affiliation not in history and not terminated:
                    history.add(affiliation)
                    entry["affiliations"].append({"name": affiliation})
        except Exception:
            pass

        stream_entry.entry = entry
        return stream_entry


class NamesServiceWriter(ServiceWriter):
    """Names service writer."""

    def __init__(self, *args, **kwargs):
        """Constructor."""
        service_or_name = kwargs.pop("service_or_name", "names")
        super().__init__(service_or_name=service_or_name, *args, **kwargs)

    def _entry_id(self, entry):
        """Get the id from an entry."""
        return entry["id"]


VOCABULARIES_DATASTREAM_READERS = {
    "orcid-http": OrcidHTTPReader,
}


VOCABULARIES_DATASTREAM_TRANSFORMERS = {
    "orcid": OrcidTransformer,
}
"""ORCiD Data Streams transformers."""


VOCABULARIES_DATASTREAM_WRITERS = {
    "names-service": NamesServiceWriter,
}
"""ORCiD Data Streams transformers."""


DATASTREAM_CONFIG = {
    "readers": [
        {
            "type": "tar",
            "args": {
                "regex": ".xml$",
            },
        },
        {"type": "xml"},
    ],
    "transformers": [{"type": "orcid"}],
    "writers": [
        {
            "type": "names-service",
            "args": {
                "identity": system_identity,
            },
        }
    ],
}
"""ORCiD Data Stream configuration.

An origin is required for the reader.
"""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Funders datastreams tests."""

from copy import deepcopy

import pytest
from invenio_access.permissions import system_identity

from invenio_vocabularies.contrib.funders.api import Funder
from invenio_vocabularies.contrib.funders.datastreams import (
    FundersServiceWriter,
    RORTransformer,
)
from invenio_vocabularies.datastreams import StreamEntry
from invenio_vocabularies.datastreams.errors import WriterError


@pytest.fixture(scope="module")
def dict_ror_entry():
    return StreamEntry(
        {
            "id": "https://ror.org/0aaaaaa11",
            "name": "Funder",
            "types": ["Facility"],
            "links": ["http://test.com"],
            "aliases": [],
            "acronyms": ["FND"],
            "status": "active",
            "wikipedia_url": "https://en.wikipedia.org/wiki/FUNDER_TEST",
            "labels": [
                {"label": "Funder", "iso639": "en"},
                {"label": "Geldgeber", "iso639": "de"},
            ],
            "email_address": None,
            "ip_addresses": [],
            "established": 1954,
            "country": {"country_code": "GR", "country_name": "Greece"},
            "relationships": [],
            "external_ids": {
                "ISNI": {"preferred": None, "all": ["0000 0001 2156 142X"]},
                "GRID": {"preferred": "grid.9132.9", "all": "grid.9132.9"},
                "FundRef": {"preferred": "000000000000", "all": []},
            },
        }
    )


@pytest.fixture(scope="module")
def expected_from_ror_json():
    return {
        "id": "0aaaaaa11",
        "name": "Funder",
        "title": {"en": "Funder", "de": "Geldgeber"},
        "country": "GR",
        "identifiers": [
            {"scheme": "ror", "identifier": "0aaaaaa11"},
            {"scheme": "isni", "identifier": "0000 0001 2156 142X"},
            {"scheme": "grid", "identifier": "grid.9132.9"},
            {"scheme": "doi", "identifier": "10.13039/000000000000"},
        ],
    }


def test_ror_transformer(app, dict_ror_entry, expected_from_ror_json):
    transformer = RORTransformer()
    assert expected_from_ror_json == transformer.apply(dict_ror_entry).entry


def test_funders_service_writer_create(app, search_clear, funder_full_data):
    writer = FundersServiceWriter()
    funder_rec = writer.write(StreamEntry(funder_full_data))
    funder_dict = funder_rec.entry.to_dict()
    assert dict(funder_dict, **funder_full_data) == funder_dict

    # not-ideal cleanup
    funder_rec.entry._record.delete(force=True)


def test_funders_service_writer_duplicate(app, search_clear, funder_full_data):
    writer = FundersServiceWriter()
    funder_rec = writer.write(stream_entry=StreamEntry(funder_full_data))
    Funder.index.refresh()  # refresh index to make changes live
    with pytest.raises(WriterError) as err:
        writer.write(stream_entry=StreamEntry(funder_full_data))

    expected_error = [f"Vocabulary entry already exists: {funder_full_data}"]
    assert expected_error in err.value.args

    # not-ideal cleanup
    funder_rec.entry._record.delete(force=True)


def test_funders_service_writer_update_existing(
    app, search_clear, funder_full_data, service
):
    # create vocabulary
    writer = FundersServiceWriter(update=True)
    orig_funder_rec = writer.write(stream_entry=StreamEntry(funder_full_data))
    Funder.index.refresh()  # refresh index to make changes live
    # update vocabulary
    updated_funder = deepcopy(funder_full_data)
    updated_funder["name"] = "Updated Name"
    # check changes vocabulary
    _ = writer.write(stream_entry=StreamEntry(updated_funder))
    funder_rec = service.read(system_identity, orig_funder_rec.entry.id)
    funder_dict = funder_rec.to_dict()

    # needed while the writer resolves from ES
    assert _.entry.id == orig_funder_rec.entry.id
    assert dict(funder_dict, **updated_funder) == funder_dict

    # not-ideal cleanup
    funder_rec._record.delete(force=True)


def test_funders_service_writer_update_non_existing(
    app, search_clear, funder_full_data, service
):
    # vocabulary item not created, call update directly
    updated_funder = deepcopy(funder_full_data)
    updated_funder["name"] = "New name"
    # check changes vocabulary
    writer = FundersServiceWriter(update=True)
    funder_rec = writer.write(stream_entry=StreamEntry(updated_funder))
    funder_rec = service.read(system_identity, funder_rec.entry.id)
    funder_dict = funder_rec.to_dict()

    assert dict(funder_dict, **updated_funder) == funder_dict

    # not-ideal cleanup
    funder_rec._record.delete(force=True)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Records-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Relations system fields."""

from flask import current_app
from invenio_records.systemfields import RelationsField
from werkzeug.utils import cached_property

from ..api import Vocabulary


class CustomFieldsRelation(RelationsField):
    """Relation field to manage custom fields.

    Iterates through all configured custom fields and collects the ones
    defining a relation dependency e.g vocabularies.
    """

    def __init__(self, fields_var):
        """Initialize the field."""
        super().__init__()
        self._fields_var = fields_var

    @cached_property
    def _fields(self):
        """Loads custom fields relations from config."""
        custom_fields = current_app.config.get(self._fields_var, {})

        relations = {}
        for cf in custom_fields:
            if getattr(cf, "relation_cls", None):
                relations[cf.name] = cf.relation_cls(
                    f"custom_fields.{cf.name}",
                    keys=cf.field_keys,
                    pid_field=Vocabulary.pid.with_type_ctx(cf.vocabulary_id),
                    cache_key=cf.vocabulary_id,
                )

        return relations

    def __set__(self, instance, values):
        """Setting the attribute."""
        raise ValueError(
            f"This field can only be set through config ({self._fields_var})"
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Names services."""

from invenio_pidstore.errors import PIDDoesNotExistError
from invenio_search.engine import dsl

from .names import record_type

NamesServiceConfig = record_type.service_config_cls


class NamesService(record_type.service_cls):
    """Name service."""

    def resolve(self, identity, id_, id_type):
        """Get the record with a given identifier.

        This method assumes that the are no duplicates in the system
        (i.e. only one name record can have a pair of identifier:scheme).
        """
        search_query = dsl.Q(
            "bool",
            minimum_should_match=1,
            must=[
                dsl.Q("term", identifiers__identifier=id_),
                dsl.Q("term", identifiers__scheme=id_type),
            ],
        )

        # max_records = 1, we assume there cannot be duplicates
        # the loading process needs to make sure of that
        results = self._read_many(identity, search_query, max_records=1)
        # cant use the results_item because it returns dicts intead of records
        total = results.hits.total["value"]
        if total == 0:
            # Not a PID but trated as such
            raise PIDDoesNotExistError(pid_type=id_type, pid_value=id_)

        # (0 < #hits <= max_records) = 1
        record = self.record_cls.loads(results[0].to_dict())
        self.require_permission(identity, "read", record=record)

        return self.result_item(
            self,
            identity,
            record,
            links_tpl=self.links_item_tpl,
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Affiliations JSONSchema tests."""

import pytest
from jsonschema.exceptions import ValidationError

from invenio_vocabularies.records.api import Vocabulary


@pytest.fixture(scope="function")
def vocabulary_json():
    """Returns the schema location."""
    return {
        "$schema": "local://vocabularies/vocabulary-v1.0.0.json",
        "description": {
            "en": "Test affiliation description",
            "es": "Descripcion de una afiliacion de test",
        },
        "icon": "test.png",
        "id": "voc-1",
        "pid": {"pk": 1, "status": "R", "pid_type": "vocid", "obj_type": "voc"},
        "props": {"example": "test"},
        "tags": ["one_tag", "two_tag"],
        "title": {"en": "Test affiliation", "es": "Afiliacion de test"},
        "type": {"id": "vocabulary", "pid_type": "vocid"},
    }


def validates(data):
    """Validates affiliation data."""
    Vocabulary(data).validate()

    return True


def fails(data):
    """Validates affiliation data."""
    pytest.raises(ValidationError, validates, data)
    return True


def test_valid_full(appctx, vocabulary_json):
    assert validates(vocabulary_json)


def test_valid_empty(appctx):
    # check there are no requirements at JSONSchema level
    data = {"$schema": "local://vocabularies/vocabulary-v1.0.0.json"}

    assert validates(data)


def test_fails_description(appctx, vocabulary_json):
    vocabulary_json["description"] = [
        "Test affiliation description",
        "Descripcion de una afiliacion de test",
    ]

    assert fails(vocabulary_json)

    vocabulary_json["description"] = [
        {"en": "Test affiliation description"},
        {"es": "Descripcion de una afiliacion de test"},
    ]

    assert fails(vocabulary_json)

    vocabulary_json["description"] = "Test affiliation description"

    assert fails(vocabulary_json)


def test_fails_icon(appctx, vocabulary_json):
    vocabulary_json["icon"] = 123
    assert fails(vocabulary_json)


def test_fails_props(appctx, vocabulary_json):
    vocabulary_json["props"] = {"key": 1234}
    assert fails(vocabulary_json)


def test_fails_tags(appctx, vocabulary_json):
    vocabulary_json["tags"] = "tag, tag too"

    assert fails(vocabulary_json)


def test_fails_title(appctx, vocabulary_json):
    vocabulary_json["title"] = ["Test affiliation", "Afiliacion de test"]

    assert fails(vocabulary_json)

    vocabulary_json["title"] = [
        {"en": "Test affiliation"},
        {"es": "Afiliacion de test"},
    ]

    assert fails(vocabulary_json)

    vocabulary_json["title"] = "Test affiliation"

    assert fails(vocabulary_json)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary names."""

from invenio_db import db
from invenio_records.dumpers import SearchDumper
from invenio_records.dumpers.indexedat import IndexedAtDumperExt
from invenio_records.dumpers.relations import RelationDumperExt
from invenio_records.systemfields import RelationsField
from invenio_records_resources.factories.factory import RecordTypeFactory
from invenio_records_resources.records.systemfields import (
    ModelPIDField,
    PIDListRelation,
)

from ...services.permissions import PermissionPolicy
from ..affiliations.api import Affiliation
from .config import NamesSearchOptions, service_components
from .schema import NameSchema

name_relations = RelationsField(
    affiliations=PIDListRelation(
        "affiliations",
        keys=["name"],
        pid_field=Affiliation.pid,
        cache_key="affiliations",
    )
)

record_type = RecordTypeFactory(
    "Name",
    # Data layer
    pid_field_cls=ModelPIDField,
    pid_field_kwargs={
        "model_field_name": "pid",
    },
    model_cls_attrs={
        # cannot set to nullable=False because it would fail at
        # service level when create({}), see records-resources.
        "pid": db.Column(db.String(255), unique=True),
    },
    schema_version="1.0.0",
    schema_path="local://names/name-v1.0.0.json",
    record_relations=name_relations,
    record_dumper=SearchDumper(
        model_fields={"pid": ("id", str)},
        extensions=[
            RelationDumperExt("relations"),
            IndexedAtDumperExt(),
        ],
    ),
    # Service layer
    service_id="names",
    service_schema=NameSchema,
    search_options=NamesSearchOptions,
    service_components=service_components,
    permission_policy_cls=PermissionPolicy,
    # Resource layer
    endpoint_route="/names",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary service schema."""

from invenio_i18n import lazy_gettext as _
from invenio_records_resources.services.records.schema import BaseRecordSchema
from marshmallow import (
    Schema,
    ValidationError,
    fields,
    post_load,
    pre_dump,
    pre_load,
    validate,
    validates_schema,
)
from marshmallow_utils.fields import SanitizedUnicode

i18n_strings = fields.Dict(
    allow_none=False,
    keys=fields.Str(validate=validate.Regexp("^[a-z]{2}$")),
    values=fields.Str(),
)
"""Field definition for language aware strings."""


class BaseVocabularyRelationSchema(Schema):
    """Base Vocabulary relation schema."""

    id = SanitizedUnicode(required=True)


class VocabularyRelationSchema(BaseVocabularyRelationSchema):
    """Vocabulary relation schema."""

    title = fields.Dict(dump_only=True)

    @pre_load
    def clean(self, data, **kwargs):
        """Removes dump_only fields.

        Why: We want to allow the output of a Schema dump, to be a valid input
             to a Schema load without causing strange issues.
        """
        value_is_dict = isinstance(data, dict)
        if value_is_dict:
            for name, field in self.fields.items():
                if field.dump_only:
                    data.pop(name, None)
        return data


class ContribVocabularyRelationSchema(Schema):
    """Base Vocabulary relation schema."""

    id = SanitizedUnicode()
    ftf_name = None  # free text field name
    parent_field_name = None

    @validates_schema
    def validate_relation_schema(self, data, **kwargs):
        """Validates that either id either the free text field are present."""
        id_ = data.get("id")
        free_text = data.get(self.ftf_name)
        if id_:
            data = {"id": id_}
        elif free_text:
            data = {self.ftf_name: free_text}

        if not id_ and not free_text:
            raise ValidationError(
                _(
                    "An existing id or a free text {ftf_name} must be present.".format(
                        ftf_name=self.ftf_name
                    )
                ),
                self.parent_field_name,
            )


class BaseVocabularySchema(BaseRecordSchema):
    """Base schema for vocabularies."""

    title = i18n_strings
    description = i18n_strings
    icon = fields.Str(allow_none=False)


class VocabularySchema(BaseVocabularySchema):
    """Service schema for vocabulary records."""

    props = fields.Dict(allow_none=False, keys=fields.Str(), values=fields.Str())
    tags = fields.List(SanitizedUnicode())
    type = fields.Str(attribute="type.id", required=True)


class ModePIDFieldVocabularyMixin:
    """Mixin for vocabularies using a model field for their PID."""

    @validates_schema
    def validate_id(self, data, **kwargs):
        """Validates ID."""
        is_create = "record" not in self.context
        if is_create and "id" not in data:
            raise ValidationError(_("Missing PID."), "id")
        if not is_create:
            data.pop("id", None)

    @post_load(pass_many=False)
    def move_id(self, data, **kwargs):
        """Moves id to pid."""
        if "id" in data:
            data["pid"] = data.pop("id")
        return data

    @pre_dump(pass_many=False)
    def extract_pid_value(self, data, **kwargs):
        """Extracts the PID value."""
        data["id"] = data.pid.pid_value
        return data


class DatastreamObject(Schema):
    """Datastream object (reader, transformer, writer)."""

    type = fields.Str(required=True)
    args = fields.Dict(keys=fields.Str(), values=fields.Field)


class TaskSchema(Schema):
    """Service schema for vocabulary tasks."""

    readers = fields.List(
        fields.Nested(DatastreamObject),
        validate=validate.Length(min=1),
        required=True,
    )
    transformers = fields.List(
        fields.Nested(DatastreamObject),
        validate=validate.Length(min=1),
        required=False,
    )
    writers = fields.List(
        fields.Nested(DatastreamObject),
        validate=validate.Length(min=1),
        required=True,
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary awards configuration."""

from flask import current_app
from invenio_i18n import lazy_gettext as _
from invenio_records_resources.services import SearchOptions
from invenio_records_resources.services.records.components import (
    DataComponent,
    RelationsComponent,
)
from invenio_records_resources.services.records.facets import TermsFacet
from invenio_records_resources.services.records.params import SuggestQueryParser
from werkzeug.local import LocalProxy

from ...services.components import ModelPIDComponent
from ..funders.facets import FundersLabels

award_schemes = LocalProxy(lambda: current_app.config["VOCABULARIES_AWARD_SCHEMES"])

awards_openaire_funders_mapping = LocalProxy(
    lambda: current_app.config["VOCABULARIES_AWARDS_OPENAIRE_FUNDERS"]
)

awards_ec_ror_id = LocalProxy(
    lambda: current_app.config["VOCABULARIES_AWARDS_EC_ROR_ID"]
)


class AwardsSearchOptions(SearchOptions):
    """Search options."""

    suggest_parser_cls = SuggestQueryParser.factory(
        fields=[
            "acronym^100",
            "title.*^50",
            "title.*._2gram",
            "title.*._3gram",
            "number^10",
        ],
    )

    facets = {
        "funders": TermsFacet(
            field="funder.id", label=_("Funders"), value_labels=FundersLabels("funders")
        )
    }


service_components = [
    # Order of components are important!
    DataComponent,
    ModelPIDComponent,
    RelationsComponent,
]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Data Streams readers tests."""
import json
import tarfile
from pathlib import Path

import pytest
import yaml

from invenio_vocabularies.datastreams.readers import (
    JsonReader,
    TarReader,
    YamlReader,
    ZipReader,
)


@pytest.fixture(scope="module")
def expected_from_yaml():
    return [{"test": {"inner": "value"}}, {"test": {"inner": "value"}}]


@pytest.fixture(scope="function")
def yaml_file(expected_from_yaml):
    filename = Path("reader_test.yaml")
    with open(filename, "w") as file:
        yaml.dump(expected_from_yaml, file)

    yield filename

    filename.unlink()  # delete created file


def test_yaml_reader(yaml_file, expected_from_yaml):
    reader = YamlReader(yaml_file)

    for idx, data in enumerate(reader.read()):
        assert data == expected_from_yaml[idx]


@pytest.fixture(scope="module")
def expected_from_tar():
    return {"test": {"inner": "value"}}


@pytest.fixture(scope="function")
def tar_file(expected_from_tar):
    """Creates a Tar file with three files (two yaml) inside.

    Each iteration should return the content of one yaml file,
    it should ignore the .other file.
    """
    files = ["file_one.yaml", "file_two.yaml", "file_three.other"]
    filename = Path("reader_test.tar.gz")
    with tarfile.open(filename, "w:gz") as tar:
        for file_ in files:
            inner_filename = Path(file_)
            with open(inner_filename, "w") as file:
                yaml.dump(expected_from_tar, file)
                tar.add(inner_filename)
            inner_filename.unlink()

    yield filename

    filename.unlink()  # delete created file


def test_tar_reader(tar_file, expected_from_tar):
    reader = TarReader(tar_file, regex=".yaml$")

    total = 0
    for data in reader.read():
        assert yaml.safe_load(data) == expected_from_tar
        total += 1

    assert total == 2  # ignored the `.other` file


def test_zip_reader(zip_file, json_list):
    reader = ZipReader(zip_file, regex=".json$")
    total = 0
    for data in reader.read():
        assert json.load(data) == json_list
        total += 1

    assert total == 2  # ignored the `.other` file


@pytest.fixture(scope="function")
def json_list_file(json_list):
    """Creates a JSON file with an array inside."""
    filename = Path("reader_test.json")
    with open(filename, mode="w") as file:
        json.dump(json_list, file)

    yield filename

    filename.unlink()  # delete created file


def test_json_list_reader(json_list_file, json_element):
    reader = JsonReader(json_list_file, regex=".json$")

    for count, data in enumerate(reader.read(), start=1):
        assert data == json_element

    assert count == 2


@pytest.fixture(scope="function")
def json_element_file(json_element):
    """Creates a JSON file with only one element inside."""
    filename = Path("reader_test.json")
    with open(filename, mode="w") as file:
        json.dump(json_element, file)
    yield filename
    filename.unlink()  # delete created file


def test_json_element_reader(json_element_file, json_element):
    reader = JsonReader(json_element_file, regex=".json$")

    for count, data in enumerate(reader.read(), start=1):
        assert data == json_element

    assert count == 1


# FIXME: add test for csv reader



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Pytest configuration.

See https://pytest-invenio.readthedocs.io/ for documentation on which test
fixtures are available.
"""

import pytest
from invenio_records_resources.proxies import current_service_registry


@pytest.fixture(scope="module")
def service():
    """Names service object."""
    return current_service_registry.get("names")


@pytest.fixture(scope="module")
def affiliations_service():
    """Names service object."""
    return current_service_registry.get("affiliations")


@pytest.fixture()
def example_affiliation(db, identity, affiliations_service):
    """Example affiliation."""
    aff = affiliations_service.create(
        identity,
        {"id": "cern", "name": "European Organization for Nuclear Research"},
    )
    affiliations_service.record_cls.index.refresh()
    yield aff
    aff._record.delete(force=True)
    affiliations_service.indexer.delete(aff._record, refresh=True)
    db.session.commit()


@pytest.fixture(scope="function")
def name_full_data():
    """Full name data."""
    return {
        "id": "0000-0001-8135-3489",
        "name": "Doe, John",
        "given_name": "John",
        "family_name": "Doe",
        "identifiers": [
            {"identifier": "0000-0001-8135-3489", "scheme": "orcid"},
            {"identifier": "gnd:4079154-3", "scheme": "gnd"},
        ],
        "affiliations": [{"id": "cern"}, {"name": "CustomORG"}],
    }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Persistent identifier provider for vocabularies."""


from invenio_pidstore.models import PIDStatus
from invenio_pidstore.providers.base import BaseProvider


class VocabularyIdProvider(BaseProvider):
    """Vocabulary identifier provider.

    This PID provider requires a Vocabulary record to be passed, and relies
    on the vocabulary record having an 'id' key and a type defined.
    """

    @classmethod
    def create(cls, object_type=None, object_uuid=None, record=None, **kwargs):
        """Create a new vocabulary identifier.

        Relies on the a vocabulary record being

        Note: if the object_type and object_uuid values are passed, then the
        PID status will be automatically setted to
        :attr:`invenio_pidstore.models.PIDStatus.REGISTERED`.

        For more information about parameters,
        see :meth:`invenio_pidstore.providers.base.BaseProvider.create`.

        :param object_type: The object type. (Default: None.)
        :param object_uuid: The object identifier. (Default: None).
        :param record: A vocabulary record.
        :param kwargs: Addtional options
        :returns: A :class:`VocabularyIdProvider` instance.
        """
        assert record is not None, "Missing or invalid 'record'."
        assert "id" in record and isinstance(
            record["id"], str
        ), "Missing 'id' key in record."

        # Retrieve pid type from type.
        pid_type = record.type.pid_type
        # Retrieve pid value form record.
        pid_value = record["id"]

        # You must assign immediately.
        assert object_uuid
        assert object_type

        return super().create(
            pid_type=pid_type,
            pid_value=pid_value,
            object_type=object_type,
            object_uuid=object_uuid,
            status=PIDStatus.REGISTERED,
        )


class CustomVocabularyPIDProvider(BaseProvider):
    """Affiliations PID provider."""

    pid_type = None
    """Type of persistent identifier."""

    @classmethod
    def create(cls, object_type=None, object_uuid=None, record=None, **kwargs):
        """Create a new affiliation identifier.

        Relies on the record having already a pid_value.

        For more information about parameters,
        see :meth:`invenio_pidstore.providers.base.BaseProvider.create`.

        :param object_type: The object type. (Default: None.)
        :param object_uuid: The object identifier. (Default: None).
        :param record: An affiliation vocabulary record.
        :param kwargs: Addtional options
        :returns: A :class:`AffiliationProvider` instance.
        """
        assert record is not None, "Missing or invalid 'record'."
        assert "id" in record and isinstance(
            record["id"], str
        ), "Missing 'id' key in record."

        # Retrieve pid value form record.
        pid_value = record["id"]

        # You must assign immediately.
        assert object_uuid
        assert object_type

        return super().create(
            pid_type=cls.pid_type,
            pid_value=pid_value,
            object_type=object_type,
            object_uuid=object_uuid,
            status=PIDStatus.REGISTERED,
        )


class PIDProviderFactory:
    """Vocabulary PID provider factory."""

    @staticmethod
    def create(pid_type, base_cls=CustomVocabularyPIDProvider):
        """Returns a CustomVocabularyPIDProvider with the given PID type."""
        provider_class_attributes = {
            "pid_type": pid_type,
        }

        return type(
            "CustomVocabularyPIDProvider", (base_cls,), provider_class_attributes
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary facets."""

from flask_principal import AnonymousIdentity
from invenio_i18n.ext import current_i18n
from invenio_records_resources.proxies import current_service_registry
from invenio_records_resources.services.errors import FacetNotFoundError
from marshmallow_utils.fields.babel import gettext_from_dict
from speaklater import make_lazy_string
from sqlalchemy.exc import NoResultFound

from ..proxies import current_service


def lazy_get_label(vocab_item):
    """Lazy evaluation of a localized vocabulary label."""
    params = {"locale": current_i18n.locale, "default_locale": "en"}

    return make_lazy_string(gettext_from_dict, vocab_item, **params)


class VocabularyLabels:
    """Fetching of vocabulary labels for facets."""

    def __init__(self, vocabulary, cache=False, service_name=None, id_field="id"):
        """Initialize the labels."""
        self.vocabulary = vocabulary
        self.cache = cache
        self.fields = ["id", "title"]  # not configurable
        self.service_name = service_name
        self.id_field = id_field

    @property
    def service(self):
        """Service property.

        It is required to access the regitry lazily to avoid out of
        application context errors.
        """
        if not self.service_name:
            return current_service
        return current_service_registry.get(self.service_name)

    def _vocab_to_label(self, vocab):
        """Returns the label string for a vocabulary entry."""
        return lazy_get_label(vocab["title"])

    def __call__(self, ids):
        """Return the mapping when evaluated."""
        identity = AnonymousIdentity()
        try:
            if not self.cache:
                vocabs = self.service.read_many(
                    identity, type=self.vocabulary, ids=ids, fields=self.fields
                )
            else:
                vocabs = self.service.read_all(
                    identity, type=self.vocabulary, fields=self.fields
                )
        except NoResultFound:
            raise FacetNotFoundError(self.vocabulary)

        labels = {}
        vocab_list = list(vocabs.hits)  # the service returns a generator
        ids = set(ids)
        seen = set()
        for vocab in vocab_list:
            # cannot loop over ids because vocab id is inside each item
            if len(ids) == len(seen):
                break

            id_ = vocab[self.id_field]
            if id_ in ids:
                labels[id_] = self._vocab_to_label(vocab)
                seen.add(id_)

        return labels



#
# This file is part of Invenio.
# Copyright (C) 2023 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Add "pid" column to names and affiliations tables."""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "55a700f897b6"
down_revision = "e1146238edd3"
branch_labels = ()
depends_on = None


pids_table = sa.sql.table(
    "pidstore_pid",
    sa.sql.column("pid_type"),
    sa.sql.column("pid_value"),
    sa.sql.column("object_uuid"),
    sa.sql.column("object_type"),
)
names_table = sa.sql.table(
    "name_metadata",
    sa.sql.column("id"),
    sa.sql.column("pid"),
)
affiliations_table = sa.sql.table(
    "affiliation_metadata",
    sa.sql.column("id"),
    sa.sql.column("pid"),
)


def upgrade():
    """Upgrade database."""
    op.add_column(
        "name_metadata",
        sa.Column("pid", sa.String(255), nullable=True),
    )
    # populate from PIDStore
    op.execute(
        names_table.update()
        .where(
            names_table.c.id == pids_table.c.object_uuid,
            pids_table.c.object_type == "rec",
            pids_table.c.pid_type == "names",
        )
        .values(pid=pids_table.c.pid_value)
    )
    op.create_unique_constraint(
        op.f("uq_name_metadata_pid"),
        "name_metadata",
        ["pid"],
    )

    op.add_column(
        "affiliation_metadata",
        sa.Column("pid", sa.String(255), nullable=True),
    )
    # populate from PIDStore
    op.execute(
        affiliations_table.update()
        .where(
            affiliations_table.c.id == pids_table.c.object_uuid,
            pids_table.c.object_type == "rec",
            pids_table.c.pid_type == "aff",
        )
        .values(pid=pids_table.c.pid_value)
    )

    op.create_unique_constraint(
        op.f("uq_affiliation_metadata_pid"),
        "affiliation_metadata",
        ["pid"],
    )


def downgrade():
    """Downgrade database."""
    op.drop_constraint(
        op.f("uq_name_metadata_pid"),
        "name_metadata",
        type_="unique",
    )
    op.drop_constraint(
        op.f("uq_affiliation_metadata_pid"),
        "affiliation_metadata",
        type_="unique",
    )
    op.drop_column("name_metadata", "pid")
    op.drop_column("affiliation_metadata", "pid")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 Northwestern University.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Test the affiliation vocabulary resources."""

from .subjects import record_type

SubjectsResourceConfig = record_type.resource_config_cls
SubjectsResourceConfig.routes["item"] = "/<path:pid_value>"

SubjectsResource = record_type.resource_cls
